import contextlib
import contextvars
import logging
import random
import re
import typing

from weave.legacy.weave import weave_types as types
from weave.legacy.weave import (
    weave_internal,
    errors,
    engine_trace,
    box,
    compile_domain,
    compile_table,
    debug_compile,
    dispatch,
    gql_op_plugin,
    gql_to_weave,
    graph,
    graph_debug,
    input_provider,
    op_args,
    partial_object,
    propagate_gql_keys,
    registry_mem,
    serialize,
    stitch,
    value_or_error,
)
from weave.legacy.weave.language_features.tagging import tagged_value_type_helpers
from weave.legacy.weave.op_def import OpDef

# These call_* functions must match the actual op implementations.
# But we don't want to import the op definitions themselves here, since
# those depend on the decorators, which aren't defined in the engine.

DEBUG_COMPILE = False


def _dispatch_error_is_client_error(
    op_name: str, input_types: dict[str, types.Type]
) -> bool:
    from weave.legacy.weave.ops_domain import wbmedia

    if op_name in set(
        (
            "tag-run",
            "count",
            "file-table",
            "offset",
            "typedDict-pick",
            "list-createIndexCheckpointTag",
            "entity-name",
            "concat",
            "string-isNumeric",
            "project-artifacts",
            "panel-table",
        )
    ):
        # All the cases we've seen of this lately are clearly client errors, so
        # we'll send back a 400!
        return True
    elif op_name == "file-path" and types.optional(
        wbmedia.ImageArtifactFileRefType()
    ).assign_type(input_types["file"]):
        # You shouldn't be able to call file-path on ImageArtifactFileRef.
        return True
    return False


def _call_run_await(run_node: graph.Node) -> graph.OutputNode:
    return graph.OutputNode(
        run_node.type.output_type.output, "run-await", {"self": run_node}  # type: ignore
    )


# We don't want to import the op definitions themselves here, since
# those depend on the decorators, which aren't defined in the engine.
def _call_execute(function_node: graph.Node) -> graph.OutputNode:
    function_node_type = typing.cast(types.Function, function_node.type)
    if isinstance(function_node, graph.ConstNode) and isinstance(
        function_node.val.type, types.Function
    ):
        return _call_execute(function_node.val)
    return graph.OutputNode(
        function_node_type.output_type, "execute", {"node": function_node}
    )


def _quote_node(node: graph.Node) -> graph.Node:
    return weave_internal.const(node)


def _static_function_types(node: graph.Node) -> typing.Optional[graph.Node]:
    # This compile time transform looks at all const-functions which have 0
    # inputs (we call these static lambdas). If it is a static lambda, then we
    # want to update the output type of the inner node and the function such
    # that the rest of the system has the correct types. This is needed because
    # we do not map into static lambdas and do not compile their inner contents
    # as it should be treated as a literal, quoted graph.
    if (
        isinstance(node, graph.ConstNode)
        and isinstance(node.type, types.Function)
        and len(node.type.input_types) == 0
    ):
        inner_node = node.val
        if isinstance(inner_node, graph.OutputNode):
            compiled_node = _compile([inner_node])[0]
            return weave_internal.const(
                weave_internal.make_output_node(
                    compiled_node.type,
                    inner_node.from_op.name,
                    inner_node.from_op.inputs,
                )
            )
    return None


def _remove_optional(t: types.Type) -> types.Type:
    if types.is_optional(t):
        return types.non_none(t)
    return t


def _dispatch_map_fn_no_refine(node: graph.Node) -> typing.Optional[graph.OutputNode]:
    if isinstance(node, graph.OutputNode):
        if node.from_op.name == "tag-indexCheckpoint":
            # I'm seeing that there is no indexCheckpoint tag present
            # on types that come from WeaveJS (at least by the time we call
            # this op). Maybe a WeaveJS bug?
            # TODO
            return node
        if node.from_op.name == "file-type":
            # since we didn't refine, the input to file-type is not correct yet.
            # if its in the graph, just trust that's what we want
            # TODO: does this work for mapped case?
            return node
        from_op = node.from_op
        op = None
        try:
            op = dispatch.get_op_for_inputs(node.from_op.name, from_op.input_types)
        except errors.WeaveDispatchError as e:
            # With the new DashUI implementation, we are not allowed to refine
            # nodes on the client. As a result, it is perfectly possible to
            # construct a graph manually (as is done when generating a template
            # from a table) where the nodes are under-typed. The result of this
            # is that we will fail to dispatch in this phase which short
            # circuits and errors early. This call to
            # `registry_mem.memory_registry.get_op(node.from_op.name)` is a
            # last-resort that blindly accepts the op by name (trusting that the
            # developer client-side constructed a good graph). Importantly, if
            # this is an invalid assumption, and the graph is truly incorrect,
            # the error will be thrown in the `compile_refine_and_propagate_gql`
            # step. This just permits the compilation to continue, given that
            # the above circumstance happens more frequently now.
            try:
                op = registry_mem.memory_registry.get_op(node.from_op.name)
            except errors.WeaveMissingOpDefError:
                pass
            if op is None:
                if _dispatch_error_is_client_error(from_op.name, from_op.input_types):
                    raise errors.WeaveBadRequest(
                        "Error while dispatching (no refine phase): %s. This is most likely a client error"
                        % from_op.name
                    )
                else:
                    raise

        params = from_op.inputs
        if isinstance(op.input_type, op_args.OpNamedArgs):
            params = {
                k: n for k, n in zip(op.input_type.arg_types, from_op.inputs.values())
            }

        output_type = node.type
        # In the case where we are dispatching to a new op, we want to use the
        # new op's `unrefined_output_type_for_params` output type - rather than
        # blindly trusting the client type.
        if not node.from_op.name.startswith("local-artifact://") and (
            # Why do we have to whitelist ops? Because we may have changed the input
            # types to, e.g. dict, which pass their inputs types through to their output
            # This solution doesn't feel generic, but it solves the boostrapping refinement
            # case on the LLM monitoring board. We refine all input nodes in the document.
            # Some of the refinement, like opMap, happens on the client side, rewriting the
            # type to be a regular list instead of ArrowWeaveList. When we get back here
            # to Python, we need that type to be ArrowWeaveList so it can be passed to withColumns.
            # If we don't whitelist dict and list, we'll pass the client type through here, which
            # is List, which will cause a dispatch error when we go to dispatch withColumns
            node.from_op.name == "dict"
            or node.from_op.name == "list"
            # We have to do this check because we do want to trust the client types in
            # many cases, like we have a get() that has already been refined. If we overwrite
            # that type here it would be "Any" and we won't be able to dispatch the next op.
            or node.from_op.name != op.name
        ):
            output_type = op.unrefined_output_type_for_params(params)

        # Consider: If the UI does a `callOpVeryUnsafe`, it is possible that the
        # graph is not correctly typed. Consider checking if they type is `Any`,
        # then we may want to use the concrete output type instead.

        res = graph.OutputNode(_remove_optional(output_type), op.uri, params)
        # logging.info("Dispatched (no refine): %s -> %s", node, res.type)
        return res
    return None


def _simple_optimizations(node: graph.Node) -> typing.Optional[graph.Node]:
    # Put simple graph transforms here!
    if not isinstance(node, graph.OutputNode):
        return None
    if node.from_op.friendly_name == "merge":
        # Merging two dicts where one is empty returns the other. The frontend
        # sends this pattern down a lot right now, and it causes us to break out
        # of arrow vectorization.
        lhs, rhs = node.from_op.inputs.values()
        if (
            isinstance(lhs, graph.OutputNode)
            and lhs.from_op.friendly_name == "dict"
            and not lhs.from_op.inputs
        ):
            return rhs
        if (
            isinstance(rhs, graph.OutputNode)
            and rhs.from_op.friendly_name == "dict"
            and not rhs.from_op.inputs
        ):
            return lhs
    elif isinstance(node, graph.OutputNode) and node.from_op.name == "count":
        # When the graph is `run.history.count`, we can avoid the more costly
        # loading of all run history and instead directly fetch the history
        # count from the server by reducing to `run.historyLineCount` which
        # compiles to a single gql edge. This is particularly helpful when
        # loading StreamTables backed by runs. It helps with the client-side map
        # refinement, table row count, and with logic that is conditioned on
        # empty tables
        arr_node = node.from_op.inputs["arr"]
        if isinstance(arr_node, graph.OutputNode) and arr_node.from_op.name.startswith(
            "run-history"
        ):
            run_node = arr_node.from_op.inputs["run"]
            return graph.OutputNode(
                node.type,
                "run-historyLineCount",
                {"run": run_node},
            )
    elif isinstance(node, graph.OutputNode) and node.from_op.name == "unique":
        # When the graph is `awl.keys.flatten.unique`, the user is really asking
        # for the columnNames. This can be reduced to a simple `awl.columnNames`
        # which is extremely fast as it is simply the property types of the
        # list!
        #
        # Note: we cannot perform such optimization on pure lists because we
        # don't have a way to operate on the type of the node itself.
        arr_node = node.from_op.inputs["arr"]
        if (
            isinstance(arr_node, graph.OutputNode)
            and arr_node.from_op.name == "ArrowWeaveList-flatten"
        ):
            arr_node_2 = arr_node.from_op.inputs["arr"]
            if (
                isinstance(arr_node_2, graph.OutputNode)
                and arr_node_2.from_op.name == "ArrowWeaveListTypedDict-keys"
            ):
                awl_node = arr_node_2.from_op.inputs["self"]
                return graph.OutputNode(
                    node.type, "ArrowWeaveListTypedDict-columnNames", {"self": awl_node}
                )
    elif isinstance(node, graph.OutputNode) and node.from_op.name == "flatten":
        from weave.legacy.weave.arrow.arrow import ArrowWeaveListType

        from .ops_arrow.list_ops import _concat_output_type

        # The operation of flattening a lists of arrow weave lists is exactly equal to the far
        # more efficient, vectorized concat operation. If this is the case, use it!.
        arr_node = node.from_op.inputs["arr"]
        if types.List().assign_type(arr_node.type) and ArrowWeaveListType().assign_type(
            arr_node.type.object_type
        ):
            return graph.OutputNode(
                _concat_output_type({"arr": arr_node.type}),
                "ArrowWeaveList-concat",
                {"arr": arr_node},
            )
    elif isinstance(node, graph.OutputNode) and node.from_op.name == "concat":
        from weave.legacy.weave.arrow.arrow import ArrowWeaveListType

        from .ops_arrow.list_ops import flatten_return_type

        # The operation of concat on a awl of lists is exactly equal to the far
        # more efficient, vectorized flatten operation. If this is the case, use it!.
        arr_node = node.from_op.inputs["arr"]
        if ArrowWeaveListType().assign_type(arr_node.type) and types.List().assign_type(
            arr_node.type.object_type
        ):
            return graph.OutputNode(
                flatten_return_type({"arr": arr_node.type}),
                "ArrowWeaveList-flatten",
                {"arr": arr_node},
            )
    return None


def _required_const_input_names(node: graph.OutputNode) -> typing.Optional[list[str]]:
    res = compile_domain.required_const_input_names(node)
    if res is not None:
        return res
    return None


def _resolve_required_consts(node: graph.Node) -> typing.Optional[graph.Node]:
    # Put simple graph transforms here!
    if not isinstance(node, graph.OutputNode):
        return None
    required_const_input_names = _required_const_input_names(node)
    if required_const_input_names is None:
        return None
    new_inputs = dict(node.from_op.inputs)
    for input_name, input_node in node.from_op.inputs.items():
        if input_name in required_const_input_names:
            if not isinstance(input_node, graph.ConstNode):
                result = weave_internal.use(_compile([input_node])[0])
                new_inputs[input_name] = graph.ConstNode(input_node.type, result)
    return graph.OutputNode(
        node.type,
        node.from_op.name,
        new_inputs,
    )


def _make_auto_op_map_fn(when_type: typing.Callable[[types.Type], bool], call_op_fn):
    def fn(node: graph.Node) -> typing.Optional[graph.Node]:
        if isinstance(node, graph.OutputNode):
            node_inputs = node.from_op.inputs
            op_def = registry_mem.memory_registry.get_op(node.from_op.name)
            if (
                op_def.name == "tag-indexCheckpoint"
                # or op_def.name == "Object-__getattr__"
                or op_def.name == "set"
                # panel_scatter and panel_distribution have the incorrect
                # input types for their config arg. They should be weave.Node.
                # We need a frontend fix to handle that. For now there's a hack
                # here.
                # TODO: Fix in frontend and panel_* and remove this hack.
                or (
                    isinstance(op_def.concrete_output_type, types.Type)
                    and op_def.concrete_output_type._base_type is not None
                    and op_def.concrete_output_type._base_type.name == "Panel"
                )
            ):
                # These are supposed to be a passthrough op, we don't want to convert
                # it. TODO: Find a more general way, maybe by type inspection?
                return None
            new_inputs: dict[str, graph.Node] = {}
            swapped = False
            for k, input_node in node_inputs.items():
                actual_input_type = input_node.type
                new_inputs[k] = input_node
                if not when_type(actual_input_type):
                    continue
                if isinstance(op_def.input_type, op_args.OpNamedArgs):
                    op_input_type = op_def.input_type.arg_types[k]
                elif isinstance(op_def.input_type, op_args.OpVarArgs):
                    op_input_type = op_def.input_type.arg_type
                else:
                    raise ValueError(
                        f"Unexpected op input type {op_def.input_type} for op {op_def.name}"
                    )
                if callable(op_input_type):
                    continue
                if not when_type(op_input_type):
                    new_inputs[k] = call_op_fn(input_node)
                    swapped = True
            if swapped:
                return graph.OutputNode(node.type, node.from_op.name, new_inputs)
        return None

    return fn


def _make_inverse_auto_op_map_fn(when_type: type[types.Type], call_op_fn):
    def fn(node: graph.Node) -> typing.Optional[graph.Node]:
        if isinstance(node, graph.OutputNode):
            node_inputs = node.from_op.inputs
            op_def = registry_mem.memory_registry.get_op(node.from_op.name)
            new_inputs: dict[str, graph.Node] = {}
            for k, input_node in node_inputs.items():
                actual_input_type = input_node.type
                new_inputs[k] = input_node
                if isinstance(actual_input_type, when_type):
                    continue
                if isinstance(op_def.input_type, op_args.OpNamedArgs):
                    op_input_type = op_def.input_type.arg_types[k]
                elif isinstance(op_def.input_type, op_args.OpVarArgs):
                    op_input_type = op_def.input_type.arg_type
                else:
                    raise ValueError(
                        f"Unexpected op input type {op_def.input_type} for op {op_def.name}"
                    )
                if callable(op_input_type):
                    continue
                if isinstance(op_input_type, when_type):
                    new_inputs[k] = call_op_fn(input_node)

            return graph.OutputNode(node.type, node.from_op.name, new_inputs)
        return None

    return fn


_await_run_outputs_map_fn = _make_auto_op_map_fn(
    lambda t: isinstance(t, types.Function)
    and isinstance(t.output_type, types.RunType),
    _call_run_await,
)

_execute_nodes_map_fn = _make_auto_op_map_fn(
    lambda t: isinstance(types.split_none(t)[1], types.Function), _call_execute
)

_quote_nodes_map_fn = _make_inverse_auto_op_map_fn(types.Function, _quote_node)


def compile_apply_column_pushdown(
    leaf_nodes: list[graph.Node], on_error: graph.OnErrorFnType = None
) -> list[graph.Node]:
    # This is specific to project-runs2 (not yet used in W&B production) for now. But it
    # is a general pattern that will work for all arrow tables.

    op_names = [
        "project-runs2",
        "run-history2",
        "mapped_run-history2",
        "run-history",
        "mapped_run-history",
        "run-history3",
        "mapped_run-history3",
    ]

    if not graph.filter_nodes_full(
        leaf_nodes,
        lambda n: isinstance(n, graph.OutputNode) and n.from_op.name in op_names,
    ):
        return leaf_nodes

    p = stitch.stitch(leaf_nodes, on_error=on_error)

    def _replace_with_column_pushdown(node: graph.Node) -> graph.Node:
        if isinstance(node, graph.OutputNode) and node.from_op.name in op_names:
            forward_obj = p.get_result(node)
            run_cols = compile_table.get_projection(forward_obj)
            if node.from_op.name == "project-runs2":
                config_cols = list(run_cols.get("config", {}).keys())
                summary_cols = list(run_cols.get("summary", {}).keys())
                return graph.OutputNode(
                    node.type,
                    "project-runs2_with_columns",
                    {
                        "project": node.from_op.inputs["project"],
                        "config_cols": weave_internal.const(config_cols),
                        "summary_cols": weave_internal.const(summary_cols),
                    },
                )
            if "run-history" in node.from_op.name:
                history_cols = list(run_cols.keys())

                if len(history_cols) > 0:
                    return graph.OutputNode(
                        node.type,
                        node.from_op.name + "_with_columns",
                        {
                            "run": node.from_op.inputs["run"],
                            "history_cols": weave_internal.const(
                                list(set([*history_cols, "_step"]))
                            ),
                        },
                    )
        return node

    return graph.map_nodes_full(leaf_nodes, _replace_with_column_pushdown, on_error)


def compile_dedupe(
    leaf_nodes: list[graph.Node], on_error: graph.OnErrorFnType = None
) -> list[graph.Node]:
    nodes: dict[str, graph.Node] = {}

    def _dedupe(node: graph.Node) -> graph.Node:
        from weave.legacy.weave import serialize

        node_id = serialize.node_id(node)
        if node_id in nodes:
            return nodes[node_id]
        nodes[node_id] = node
        return node

    return graph.map_nodes_full(leaf_nodes, _dedupe, on_error)


def compile_fix_calls(
    nodes: typing.List[graph.Node],
    on_error: graph.OnErrorFnType = None,
) -> typing.List[graph.Node]:
    return graph.map_nodes_full(nodes, _dispatch_map_fn_no_refine, on_error)


def compile_simple_optimizations(
    nodes: typing.List[graph.Node],
    on_error: graph.OnErrorFnType = None,
) -> typing.List[graph.Node]:
    return graph.map_nodes_full(nodes, _simple_optimizations, on_error)


def compile_resolve_required_consts(
    nodes: typing.List[graph.Node],
    on_error: graph.OnErrorFnType = None,
) -> typing.List[graph.Node]:
    return graph.map_nodes_full(nodes, _resolve_required_consts, on_error)


def compile_await(
    nodes: typing.List[graph.Node],
    on_error: graph.OnErrorFnType = None,
) -> typing.List[graph.Node]:
    return graph.map_nodes_full(nodes, _await_run_outputs_map_fn, on_error)


def compile_execute(
    nodes: typing.List[graph.Node],
    on_error: graph.OnErrorFnType = None,
) -> typing.List[graph.Node]:
    # Actually does the execution here in compile phase.
    # I made this change to handle cases where we need to pass mutations through
    # execute calls, which happens when we have Nodes stored in Const nodes (panels)
    # that we are mutating.
    # However I think I later solved this with client-side execution and it can maybe
    # be removed.
    # I'm leaving this for now as it doesn't affect W&B prod (which never calls execute).
    with_execute_ops = graph.map_nodes_full(nodes, _execute_nodes_map_fn, on_error)
    return with_execute_ops

    def _replace_execute(node: graph.Node) -> typing.Optional[graph.Node]:
        if isinstance(node, graph.OutputNode) and node.from_op.name == "execute":
            res = weave_internal.use(node.from_op.inputs["node"])
            if not isinstance(res, graph.Node):
                raise ValueError(
                    f"Expected node to be a Node, got {res} of type {type(res)}"
                )
            return compile_fix_calls([res])[0]
        return None

    return graph.map_nodes_full(with_execute_ops, _replace_execute)


def _resolve_function_calls(node: graph.Node) -> typing.Optional[graph.Node]:
    if (
        not isinstance(node, graph.OutputNode)
        or node.from_op.name != "function-__call__"
    ):
        return node

    inputs = list(node.from_op.inputs.values())
    fn_node = inputs[0]
    if not (
        isinstance(fn_node, graph.ConstNode)
        and isinstance(fn_node.type, types.Function)
    ):
        return node

    while isinstance(fn_node.val, graph.ConstNode) and isinstance(
        fn_node.type, types.Function
    ):
        fn_node = fn_node.val

    return weave_internal.better_call_fn(fn_node, *inputs[1:])


def compile_function_calls(
    nodes: typing.List[graph.Node],
    on_error: graph.OnErrorFnType = None,
) -> typing.List[graph.Node]:
    return graph.map_nodes_full(nodes, _resolve_function_calls, on_error)


def compile_quote(
    nodes: typing.List[graph.Node],
    on_error: graph.OnErrorFnType = None,
) -> typing.List[graph.Node]:
    return graph.map_nodes_full(nodes, _quote_nodes_map_fn, on_error)


def compile_static_function_types(
    nodes: typing.List[graph.Node],
    on_error: graph.OnErrorFnType = None,
) -> typing.List[graph.Node]:
    return graph.map_nodes_full(nodes, _static_function_types, on_error)


def _needs_gql_propagation(node: graph.OutputNode) -> bool:
    """Determines if a node needs to propagate the GQL keys of its inputs."""
    fq_opname = node.from_op.full_name
    opdef = registry_mem.memory_registry.get_op(fq_opname)
    plugin = gql_op_plugin.get_gql_plugin(opdef)
    first_arg_name = gql_op_plugin.first_arg_name(opdef)

    if first_arg_name is None:
        return False

    first_arg_type: types.Type = node.from_op.inputs[first_arg_name].type
    unwrapped_first_arg_type, _ = tagged_value_type_helpers.unwrap_tags(first_arg_type)

    if opdef.derived_from and opdef.derived_from.derived_ops["mapped"] == opdef:
        plugin = gql_op_plugin.get_gql_plugin(opdef.derived_from)
        unwrapped_first_arg_type = typing.cast(
            types.List, unwrapped_first_arg_type
        ).object_type

    return opdef.is_gql_root_resolver() or (
        plugin is not None
        and plugin.gql_op_output_type is not None
        and isinstance(unwrapped_first_arg_type, partial_object.PartialObjectType)
    )


def _initialize_gql_types_map_fn(node: graph.Node) -> typing.Optional[graph.Node]:
    if isinstance(node, graph.OutputNode):
        from_op = node.from_op

        if from_op.name == "gqlroot-wbgqlquery":
            # get the initial type
            assert "query_str" in from_op.inputs and isinstance(
                from_op.inputs["query_str"], graph.ConstNode
            )

            output_type = gql_to_weave.get_query_weave_type(
                compile_domain.normalize_gql_query_string(
                    from_op.inputs["query_str"].val
                )
            )

            return graph.OutputNode(
                output_type,
                "gqlroot-wbgqlquery",
                {
                    **from_op.inputs,
                    "output_type": graph.ConstNode(types.TypeType(), output_type),
                },
            )

        if from_op.name == "gqlroot-querytoobj":
            assert "gql_query_fragment" in from_op.inputs and isinstance(
                from_op.inputs["gql_query_fragment"], graph.ConstNode
            )
            inner_fragment = from_op.inputs["gql_query_fragment"].val

            assert "output_type" in from_op.inputs and isinstance(
                from_op.inputs["output_type"], graph.ConstNode
            )
            output_type = from_op.inputs["output_type"].val

            if isinstance(output_type, partial_object.PartialObjectTypeGeneratorType):
                key_type = typing.cast(
                    types.TypedDict,
                    gql_to_weave.get_query_weave_type(
                        compile_domain.normalize_gql_query_string(
                            compile_domain.fragment_to_query(inner_fragment)
                        )
                    ),
                )

                key = gql_to_weave.get_outermost_alias(inner_fragment)
                subtype = typing.cast(types.TypedDict, key_type.property_types[key])
                output_type = output_type.with_keys(subtype.property_types)

                return graph.OutputNode(
                    output_type,
                    "gqlroot-querytoobj",
                    {
                        **from_op.inputs,
                        "output_type": graph.ConstNode(types.TypeType(), output_type),
                    },
                )

    return node


def compile_initialize_gql_types(
    nodes: typing.List[graph.Node],
    on_error: graph.OnErrorFnType = None,
) -> typing.List[graph.Node]:
    return graph.map_nodes_full(nodes, _initialize_gql_types_map_fn, on_error)


def _call_gql_propagate_keys(
    node: graph.OutputNode, p: stitch.StitchedGraph, original_node: graph.Node
) -> types.Type:
    """Calls the GQL key propagation function for a node."""

    const_node_input_vals = {
        key: value.val
        for key, value in node.from_op.inputs.items()
        if isinstance(value, graph.ConstNode)
    }
    ip = input_provider.InputAndStitchProvider(
        const_node_input_vals, p.get_result(original_node)
    )

    # Propagate GQL types
    return propagate_gql_keys.propagate_gql_keys(node, ip)


def compile_refine_and_propagate_gql(
    nodes: typing.List[graph.Node],
    on_error: graph.OnErrorFnType = None,
) -> typing.List[graph.Node]:
    return _compile_refine_and_propagate_gql_inner(nodes, False, on_error)


def _compile_refine_and_propagate_gql_inner(
    nodes: typing.List[graph.Node],
    skip_stitch: bool = False,
    on_error: graph.OnErrorFnType = None,
) -> typing.List[graph.Node]:
    # Stitch is needed for gql key propagation only. If we try to stitch
    # a graph that does not have gqlroot-wbgqlquery, it may fail completely.
    # So we only stitch if we have a gqlroot-wbgqlquery node.

    use_stitch = not skip_stitch and (
        len(
            graph.filter_nodes_full(
                nodes,
                lambda n: isinstance(n, graph.OutputNode)
                and n.from_op.name == "gqlroot-wbgqlquery",
            )
        )
        > 0
    )

    # This lets us maintain a reverse mapping from the new nodes that are created
    # in the graph.map_nodes call to _dispatch_map_fn_refining to the original
    # nodes that were used in stitch. It's a bit hacky, but it allows us to store
    # the original nodes in the pre-dispatch graph as a list, then use a nonlocal
    # index to access them in _dispatch_map_fn_refining.
    p = stitch.stitch(nodes) if use_stitch else None
    node_array: list[graph.Node] = []

    def _ident(node: graph.Node) -> graph.Node:
        node_array.append(node)
        return node

    graph.map_nodes_full(nodes, _ident, on_error)

    i = -1

    def _dispatch_map_fn_refining(
        node: graph.Node,
    ) -> typing.Optional[graph.OutputNode]:
        # give us access to the index variable declared above
        nonlocal i
        i += 1

        if isinstance(node, graph.OutputNode):
            from_op = node.from_op

            if from_op.name == "gqlroot-wbgqlquery":
                # We skip the refine for this special node because we have already determined its type during
                # the initialize_gql_types phase, so it is already "refined". The lazy_call() below would wipe
                # that type out, leading to gql types not being propagated at all. So we skip it and use the
                # correct, previously calculated type from that phase.

                return node

            try:
                op = dispatch.get_op_for_inputs(from_op.name, from_op.input_types)
                params = from_op.inputs
                if isinstance(op.input_type, op_args.OpNamedArgs):
                    params = {
                        k: n
                        for k, n in zip(
                            op.input_type.arg_types, from_op.inputs.values()
                        )
                    }

                fixed_params = _propagate_updated_types_through_lambdas(
                    op, params, on_error
                )
                res = op.lazy_call(**fixed_params)

                # The GQL key propagation logic needs to happen in the refine pass rather than the GQL
                # compile pass. This is because the gql_op_output_types need refined input types or else
                # they can produce incorrect results. For example, consider this node:
                #
                #   root-project(...).filteredRuns(...).limit(1).run-summary()
                #                            .pick(my_table).file-table()
                #                            .table-rows().dropna().concat()
                #
                # During GQL compilation, the unrefined type of run-summary is:
                #
                #   TaggedValueType({...},
                #     List(object_type=TaggedValueType({['run']}, Dict(key_type=String(), object_type=Any()))))
                #
                # If we encounter ops without a gql_op_output_type, like pick(), we need to call
                # unrefined_output_type_for_params(). This can lead to invalid unrefined types propagating.
                #
                # In this example, after pick() and dropna() we end up with an invalid list type.
                # When concat() then calls its unrefined_output_type_for_params(), it errors because
                # we've passed an invalid list type to it.
                #
                # By propagating keys after refine, we avoid these unrefined type issues. Each
                # node's type has been refined before we propagate keys, so instead of Dict(String(), Any()),
                # we get the actual refined typedDict type.

                # We can't do GQL key propagation as a separate post-refine step today, because doing so
                # would that would re-run the whole graph if
                # nodes are modified after refine. We avoid this re-execution by doing propagation during
                # refine. In the future, if we can call compile_refine without triggering re-execution,
                # we could move this to a separate post-refine step.

                # We need the special gql_op_output_type instead of using callable output types because the
                # gql_op_output_type needs access to an InputProvider to traverse key trees and generate
                # aliases. Normal callable output types don't receive this.

                # Overall, propagating GQL keys during refine simplifies the logic by avoiding issues
                # around unrefined types and triggering re-execution when compared against doing it during the
                # compile_gql phase. However, it does couple the key propagation to the refine pass. Future
                # refactors could aim to decouple this, if we can avoid triggering re-execution.

                if p is not None and _needs_gql_propagation(res):
                    res.type = _call_gql_propagate_keys(res, p, node_array[i])

                # logging.info("Dispatched (refine): %s -> %s", node, res.type)
                return res
            except errors.WeaveDispatchError:
                logging.error(
                    "Error while dispatching (refine phase)\n!=!=!=!=!\nName: %s\nInput Types: %s\nExpression: %s",
                    from_op.name,
                    from_op.input_types,
                    re.sub(r'[\\]+"', '"', graph_debug.node_expr_str_full(node)),
                )
                if _dispatch_error_is_client_error(from_op.name, from_op.input_types):
                    err = errors.WeaveBadRequest(
                        "Error while dispatching: %s. This is most likely a client error"
                        % from_op.name
                    )
                    err.fingerprint = [
                        "error-while-dispatching-client-error",
                        from_op.name,
                    ]
                    raise err
                else:
                    raise
            except:
                logging.error(
                    "Error while dispatching (refine phase)\n!=!=!=!=!\nName: %s\nInput Types: %s\nExpression: %s",
                    from_op.name,
                    from_op.input_types,
                    re.sub(r'[\\]+"', '"', graph_debug.node_expr_str_full(node)),
                )
                raise

        return None

    return graph.map_nodes_full(nodes, _dispatch_map_fn_refining, on_error)


def _propagate_updated_types_through_lambdas(
    op: OpDef, params: dict[str, graph.Node], on_error: graph.OnErrorFnType = None
):
    """
    This method ensures that the lambda variables for a given op are correctly
    typed. Since the lambda function's variable typing is dependent on: a) the
    op; b) the prior params, we can only correctly determine the type after the
    prior args are correctly typed. Moreover, we need the op itself to determine
    the expected type of the lambda variables. This function updates the lambda
    stub (type of the function itself) as well as the internal variable nodes
    accordingly. It then refines the inner lambda function to ensure that the
    updated variable types are correctly propagated through the lambda function.

    Optimizations: 1) We call `_compile_refine_and_propagate_gql_inner` inside
    of this function, which means we "compile" the inner part of the lambda twice.
    If instead, we somehow propagated the updated var types before traversing into
    them, we could do this in a single pass.
    """

    if not isinstance(op.input_type, op_args.OpNamedArgs):
        return params

    updated_params: dict[str, graph.Node] = {**params}

    for k, target_node in params.items():
        if not (
            isinstance(target_node, graph.ConstNode)
            and isinstance(target_node.type, types.Function)
        ):
            continue

        input_arg_type = op.input_type.arg_types[k]

        if not callable(input_arg_type):
            continue

        expected_type = input_arg_type({k: p.type for k, p in updated_params.items()})

        if not isinstance(expected_type, types.Function):
            continue

        updated_input_types = {**target_node.type.input_types}
        expected_input_types = expected_type.input_types
        dirty_input_keys = set()
        for input_key, expected_input_type in expected_input_types.items():
            if input_key not in updated_input_types or expected_input_type.assign_type(
                updated_input_types[input_key]
            ):
                continue
            dirty_input_keys.add(input_key)
            updated_input_types[input_key] = expected_input_types[input_key]

        if len(dirty_input_keys) == 0:
            continue

        def _update_fn_vars(node: graph.Node) -> typing.Optional[graph.Node]:
            if isinstance(node, graph.VarNode) and node.name in dirty_input_keys:
                return graph.VarNode(updated_input_types[node.name], node.name)
            return node

        working_nodes = [target_node.val]
        working_nodes = graph.map_nodes_top_level(
            [target_node.val], _update_fn_vars, on_error
        )
        working_nodes = _compile_refine_and_propagate_gql_inner(
            working_nodes, True, on_error
        )
        updated_params[k] = graph.ConstNode(
            types.Function(updated_input_types, target_node.type.output_type),
            working_nodes[0],
        )
    return updated_params


def _node_ops(node: graph.Node) -> typing.Optional[graph.Node]:
    if not isinstance(node, graph.OutputNode):
        return None

    # weavified ops are expanded here.
    op_def = registry_mem.memory_registry.get_op(node.from_op.name)
    if op_def.weave_fn is not None:
        return weave_internal.call_fn(op_def.weave_fn, node.from_op.inputs)

    # otherwise we expand this hard-coded list
    if node.from_op.name not in [
        "RunChain-history",
        "panel_table-active_data",
        "panel_table-active_row",
        "Query-selected",
        "panel_plot-selected_data",
        "panel_table-all_rows",
        "stream_table-rows",
        "panel_trace-active_span",
        "Facet-selected",
    ]:
        return None
    new_node = typing.cast(graph.Node, weave_internal.use(node))
    if new_node is None or isinstance(new_node, box.BoxedNone):
        return graph.ConstNode(types.NoneType(), None)

    # The result will typically contain expressions that were sent down as
    # part of panel configs within Const nodes. They haven't been handled
    # by deserialize/compile yet.
    new_node_fixed_calls = compile_fix_calls([new_node])[0]

    # Do it again to handle nested calls
    return compile_node_ops([new_node_fixed_calls])[0]


def compile_node_ops(
    nodes: typing.List[graph.Node],
    on_error: graph.OnErrorFnType = None,
) -> typing.List[graph.Node]:
    # Mission critical that we do not map into static lambdas. We do
    # not want to expand nodes that are meant by the caller to be interpreted
    # as their exact node.
    return graph.map_nodes_full(nodes, _node_ops, on_error)


# This compile pass using the `top_level` mapper since we recurse manually. We can't use
# the full mapper, because we need to traverse when encountering the lambda op itself,
# rather than the const node since we need uniqueness based on the lambda op.
def compile_lambda_uniqueness(
    nodes: typing.List[graph.Node],
    on_error: graph.OnErrorFnType = None,
) -> typing.List[graph.Node]:
    return graph.map_nodes_top_level(nodes, _compile_lambda_uniqueness, on_error)


def _compile_lambda_uniqueness(node: graph.Node) -> typing.Optional[graph.Node]:
    # In the case that the node has a lambda op, then we recurse into any
    # lambda parameter. If the result is a new output var, then we update the
    # current node, ensuring that every lambda parameter is unique to the op.
    #
    # There is another benefit of this approach: parts of the lambda graph which are
    # shared between lambdas AND don't have a var ancestor will be shared in memory.
    # This means that if a lambda has a shared "non-dependent" parameter, then it will
    # only be executed once, and will be treated as the same logical op in stitch operations.
    # This is the desired behavior.
    if isinstance(node, graph.OutputNode):
        new_inputs = {}
        for input_key, input_node in node.from_op.inputs.items():
            if isinstance(input_node, graph.ConstNode) and isinstance(
                input_node.val, graph.Node
            ):
                uniq_lambda = compile_lambda_uniqueness([input_node.val])[0]
                if uniq_lambda is not input_node.val:
                    new_inputs[input_key] = graph.ConstNode(
                        input_node.type, uniq_lambda
                    )
        if len(new_inputs) > 0:
            use_inputs: dict[str, graph.Node] = {
                k: new_inputs.get(k, v) for k, v in node.from_op.inputs.items()
            }
            return weave_internal.make_output_node(
                node.type, node.from_op.name, use_inputs
            )

    # This is where the magic happens. We need to ensure that var nodes
    # are unique in memory
    elif isinstance(node, graph.VarNode):
        return weave_internal.make_var_node(node.type, node.name)

    return node


const_node = weave_internal.make_const_node(types.NoneType(), None)


def _track_errors(fn):
    def final(
        nodes: typing.List[graph.Node],
    ) -> value_or_error.ValueOrErrors[graph.Node]:
        compile_errors = []

        def on_error(ndx: int, e: Exception):
            compile_errors.append((ndx, e))
            return const_node

        inner_res = fn(nodes, on_error)

        results: list[value_or_error.ValueOrError[graph.Node]] = [
            value_or_error.Value(node) for node in inner_res
        ]
        for ndx, e in compile_errors:
            results[ndx] = value_or_error.Error(e)

        return value_or_error.ValueOrErrors(results)

    return final


def _compile(
    nodes: typing.List[graph.Node],
) -> value_or_error.ValueOrErrors[graph.Node]:
    tracer = engine_trace.tracer()
    # logging.info("Starting compilation of graph with %s leaf nodes" % len(nodes))

    results = value_or_error.ValueOrErrors.from_values(nodes)

    # If we're being called from WeaveJS, we need to use dispatch to determine
    # which ops to use. Critically, this first phase does not actually refine
    # op output types, so after this, the types in the graph are not yet correct.
    with tracer.trace("compile:fix_calls"):
        results = results.batch_map(_track_errors(compile_fix_calls))

    # Auto-transforms, where we insert operations to convert between types
    # as needed.
    # TODO: is it ok to have this before final refine?
    with tracer.trace("compile:await"):
        results = results.batch_map(_track_errors(compile_await))
    with tracer.trace("compile:execute"):
        results = results.batch_map(_track_errors(compile_execute))
    with tracer.trace("compile:function_calls"):
        results = results.batch_map(_track_errors(compile_function_calls))
    with tracer.trace("compile:quote"):
        # Mission critical to call `compile:quote` before and node re-writing
        # compilers such as compile:node_ops and compile:gql. Why?:
        #
        # It is useful to define a "static lambda". A "static lambda" is a const
        # node of type function with no inputs. This is used in our system to
        # represent a constant value which is a node. Useful for generating
        # boards or any sort of op that operates on nodes themselves.
        #
        # Moreover, this compile step will automatically "quote" inputs to ops
        # that expect node inputs - effectively making static lambdas when
        # called for.
        #
        # Furthermore, it is important to know that stream table rows (and many
        # other ops) now support expansion (meaning they get expanded into a
        # chain of new nodes in the compile pass).
        #
        # Conceptually, this created an issue: Compile passes that mutate nodes
        # (eg node expansion or gql compile) would modify the quoted node. But,
        # these functions that consume nodes do not want modified nodes.
        # Instead, we want the raw node that the caller intended. For this
        # reason we should always call compile:quote before any node re-writing.
        results = results.batch_map(_track_errors(compile_quote))

    with tracer.trace("compile:static_function_types"):
        results = results.batch_map(_track_errors(compile_static_function_types))

    # Some ops require const input nodes. This pass executes any branches necessary
    # to ensure that requirement holds.
    # Only gql ops require this for now.
    with tracer.trace("compile:resolve_required_consts"):
        results = results.batch_map(_track_errors(compile_resolve_required_consts))

    with tracer.trace("compile:node_ops"):
        results = results.batch_map(_track_errors(compile_node_ops))

    with tracer.trace("compile:simple_optimizations"):
        # Simple Optimizations should happen after `node_ops` to ensure we operate on
        # the expanded nodes.
        results = results.batch_map(_track_errors(compile_simple_optimizations))

    # The node ops phase above can expand nodes, leading to new nodes in the graph
    # that are potentially duplicates of others. dedupe will merge these nodes.
    with tracer.trace("compile:dedupe"):
        results = results.batch_map(_track_errors(compile_dedupe))

    # Stitch is used in stages following this one. Stitch requires that lambdas
    # are unique in memory if they are arguments to unique ops. We can receive
    # graphs that violate this requirement, and dedupe will happily merge lambdas
    # even if they are used in different ops. lambda_uniqueness pulls them back
    # apart.
    with tracer.trace("compile:lambda_uniqueness"):
        results = results.batch_map(_track_errors(compile_lambda_uniqueness))

    # Now that we have the correct calls, we can do our forward-looking pushdown
    # optimizations. These do not depend on having correct types in the graph.

    with tracer.trace("compile:gql_query"):
        results = results.batch_map(
            _track_errors(compile_domain.apply_domain_op_gql_translation)
        )

    with tracer.trace("compile:initialize_gql_types"):
        results = results.batch_map(_track_errors(compile_initialize_gql_types))

    with tracer.trace("compile:column_pushdown"):
        results = results.batch_map(_track_errors(compile_apply_column_pushdown))

    # Final refine, to ensure the graph types are exactly what Weave python
    # produces. This phase can execute parts of the graph. It's very important
    # that this is the final phase, so that when we execute the rest of the
    # graph, we reuse any results produced in this phase, instead of re-executing
    # those nodes.
    with tracer.trace("compile:refine_and_propagate_gql"):
        results = results.batch_map(_track_errors(compile_refine_and_propagate_gql))

    # This is very expensive!
    # loggable_nodes = graph_debug.combine_common_nodes(n)
    # logging.info(
    #     "Compilation complete. Result nodes:\n%s",
    #     "\n".join(graph_debug.node_expr_str_full(n) for n in loggable_nodes),
    # )

    if DEBUG_COMPILE:
        debug_compile.check_weave0_compile_result(
            nodes,
            [item if err == None else const_node for item, err in results.iter_items()],
        )

    return results


_compile_disabled: contextvars.ContextVar[bool] = contextvars.ContextVar(
    "_compile_disabled", default=False
)


def _is_compiling() -> bool:
    return _compile_disabled.get()


@contextlib.contextmanager
def disable_compile():
    token = _compile_disabled.set(True)
    try:
        yield
    finally:
        _compile_disabled.reset(token)


@contextlib.contextmanager
def enable_compile():
    token = _compile_disabled.set(False)
    try:
        yield
    finally:
        _compile_disabled.reset(token)


def compile(
    nodes: typing.List[graph.Node],
) -> value_or_error.ValueOrErrors[graph.Node]:
    """
    This method is used to "compile" a list of nodes. Here we can add any
    optimizations or graph rewrites
    """
    # The refine phase may execute parts of the graph. Executing recursively
    # calls compile. Use context to ensure we only compile the top level
    # graph.
    if _is_compiling():
        return value_or_error.ValueOrErrors.from_values(nodes)
    with disable_compile():
        return _compile(nodes)



import typing


from weave.legacy.weave import weave_types as types
from weave.legacy.weave import gql_op_plugin, graph, input_provider, op_def, partial_object, registry_mem


def _propagate_gql_keys_for_node(
    opdef: "op_def.OpDef",
    node: graph.OutputNode,
    key_fn: gql_op_plugin.GQLOutputTypeFn,
    ip: input_provider.InputProvider,
) -> types.Type:
    # Mutates node
    # TODO: see if this can be done without mutations
    from weave.legacy.weave.language_features.tagging import (
        opdef_util,
        tagged_value_type,
        tagged_value_type_helpers,
    )

    input_types = node.from_op.input_types

    first_arg_name = gql_op_plugin.first_arg_name(opdef)
    if first_arg_name is None:
        raise ValueError("OpDef does not have named args, cannot propagate GQL keys")

    # unwrap and rewrap tags
    original_input_type = input_types[first_arg_name]

    unwrapped_input_type, wrap = tagged_value_type_helpers.unwrap_tags(
        original_input_type
    )

    is_mapped = opdef.derived_from and opdef.derived_from.derived_ops["mapped"] == opdef
    if is_mapped:
        unwrapped_input_type = typing.cast(types.List, unwrapped_input_type).object_type

    # key fn operates on untagged types
    new_output_type = key_fn(ip, unwrapped_input_type)

    if isinstance(new_output_type, types.Invalid):
        raise ValueError('GQL key function returned "Invalid" type')

    if is_mapped:
        new_output_type = types.List(new_output_type)

    # now we rewrap the types to propagate the tags
    if opdef_util.should_tag_op_def_outputs(opdef):
        new_output_type = tagged_value_type.TaggedValueType(
            types.TypedDict({first_arg_name: original_input_type}), new_output_type
        )
    elif opdef_util.should_flow_tags(opdef):
        new_output_type = wrap(new_output_type)

    return new_output_type


def propagate_gql_keys(
    maybe_refined_node: graph.OutputNode,
    ip: input_provider.InputProvider,
) -> types.Type:
    if not maybe_refined_node.from_op.name == "gqlroot-wbgqlquery":
        fq_opname = maybe_refined_node.from_op.full_name
        opdef = registry_mem.memory_registry.get_op(fq_opname)
        plugin = gql_op_plugin.get_gql_plugin(opdef)

        key_fn: typing.Optional[gql_op_plugin.GQLOutputTypeFn] = None

        if plugin is None or plugin.gql_op_output_type is None:
            if fq_opname.endswith("GQLResolver"):
                original_opdef = registry_mem.memory_registry.get_op(
                    fq_opname.replace("GQLResolver", "")
                )
                original_plugin = gql_op_plugin.get_gql_plugin(original_opdef)
                if (
                    original_plugin is not None
                    and original_plugin.gql_op_output_type is not None
                ):
                    key_fn = original_plugin.gql_op_output_type
            if opdef.derived_from and opdef.derived_from.derived_ops["mapped"] == opdef:
                scalar_plugin = gql_op_plugin.get_gql_plugin(opdef.derived_from)
                if scalar_plugin is not None:
                    key_fn = scalar_plugin.gql_op_output_type
        else:
            key_fn = plugin.gql_op_output_type

        if key_fn is not None:
            return _propagate_gql_keys_for_node(opdef, maybe_refined_node, key_fn, ip)

    return maybe_refined_node.type



import importlib
import typing
from functools import wraps
from typing import Any, Union

from typing_extensions import AsyncIterator, Iterator

import weave
from weave.trace.op_extensions.accumulator import _IteratorWrapper, add_accumulator
from weave.trace.patcher import MultiPatcher, SymbolPatcher

if typing.TYPE_CHECKING:
    from anthropic.lib.streaming import MessageStream
    from anthropic.types import Message, MessageStreamEvent


def anthropic_accumulator(
    acc: typing.Optional["Message"],
    value: "MessageStreamEvent",
) -> "Message":
    from anthropic.types import (
        ContentBlockDeltaEvent,
        Message,
        MessageDeltaEvent,
        TextBlock,
        Usage,
    )

    if acc is None:
        if hasattr(value, "message"):
            acc = Message(
                id=value.message.id,
                role=value.message.role,
                content=[],
                model=value.message.model,
                stop_reason=value.message.stop_reason,
                stop_sequence=value.message.stop_sequence,
                type=value.message.type,  # Include the type field
                usage=Usage(input_tokens=0, output_tokens=0),
            )
        else:
            raise ValueError("Initial event must contain a message")

    # Merge in the usage info if available
    if hasattr(value, "message") and value.message.usage is not None:
        acc.usage.input_tokens += value.message.usage.input_tokens

    # Accumulate the content if it's a ContentBlockDeltaEvent
    if isinstance(value, ContentBlockDeltaEvent) and hasattr(value.delta, "text"):
        if acc.content and isinstance(acc.content[-1], TextBlock):
            acc.content[-1].text += value.delta.text
        else:
            acc.content.append(TextBlock(type="text", text=value.delta.text))

    # Handle MessageDeltaEvent for stop_reason and stop_sequence
    if isinstance(value, MessageDeltaEvent):
        if hasattr(value.delta, "stop_reason") and value.delta.stop_reason:
            acc.stop_reason = value.delta.stop_reason
        if hasattr(value.delta, "stop_sequence") and value.delta.stop_sequence:
            acc.stop_sequence = value.delta.stop_sequence
        if hasattr(value, "usage") and value.usage.output_tokens:
            acc.usage.output_tokens = value.usage.output_tokens

    return acc


# Unlike other integrations, streaming is based on input flag
def should_use_accumulator(inputs: typing.Dict) -> bool:
    return isinstance(inputs, dict) and bool(inputs.get("stream"))


def create_wrapper_sync(
    name: str,
) -> typing.Callable[[typing.Callable], typing.Callable]:
    def wrapper(fn: typing.Callable) -> typing.Callable:
        "We need to do this so we can check if `stream` is used"
        op = weave.op()(fn)
        op.name = name  # type: ignore
        return add_accumulator(
            op,  # type: ignore
            make_accumulator=lambda inputs: anthropic_accumulator,
            should_accumulate=should_use_accumulator,
        )

    return wrapper


# Surprisingly, the async `client.chat.completions.create` does not pass
# `inspect.iscoroutinefunction`, so we can't dispatch on it and must write
# it manually here...
def create_wrapper_async(
    name: str,
) -> typing.Callable[[typing.Callable], typing.Callable]:
    def wrapper(fn: typing.Callable) -> typing.Callable:
        def _fn_wrapper(fn: typing.Callable) -> typing.Callable:
            @wraps(fn)
            async def _async_wrapper(
                *args: typing.Any, **kwargs: typing.Any
            ) -> typing.Any:
                return await fn(*args, **kwargs)

            return _async_wrapper

        "We need to do this so we can check if `stream` is used"
        op = weave.op()(_fn_wrapper(fn))
        op.name = name  # type: ignore
        return add_accumulator(
            op,  # type: ignore
            make_accumulator=lambda inputs: anthropic_accumulator,
            should_accumulate=should_use_accumulator,
        )

    return wrapper


## This part of the code is for dealing with the other way
## of streaming, by calling Messages.stream
## this has 2 options: event based or text based.
## This code handles both cases by patching the _IteratorWrapper
## and adding a text_stream property to it.


def anthropic_stream_accumulator(
    acc: typing.Optional["Message"],
    value: "MessageStream",
) -> "Message":
    from anthropic.lib.streaming._types import MessageStopEvent

    if acc is None:
        acc = ""
    if isinstance(value, MessageStopEvent):
        acc = value.message
    return acc


class AnthropicIteratorWrapper(_IteratorWrapper):
    def __getattr__(self, name: str) -> Any:
        """Delegate all other attributes to the wrapped iterator."""
        if name in [
            "_iterator_or_ctx_manager",
            "_on_yield",
            "_on_error",
            "_on_close",
            "_on_finished_called",
            "_call_on_error_once",
            "text_stream",
        ]:
            return object.__getattribute__(self, name)
        return getattr(self._iterator_or_ctx_manager, name)

    def __stream_text__(self) -> Union[Iterator[str], AsyncIterator[str]]:
        if isinstance(self._iterator_or_ctx_manager, AsyncIterator):
            return self.__async_stream_text__()
        else:
            return self.__sync_stream_text__()

    def __sync_stream_text__(self) -> Iterator[str]:  # type: ignore
        for chunk in self:  # type: ignore
            if chunk.type == "content_block_delta" and chunk.delta.type == "text_delta":  # type: ignore
                yield chunk.delta.text  # type: ignore

    async def __async_stream_text__(self) -> AsyncIterator[str]:  # type: ignore
        async for chunk in self:  # type: ignore
            if chunk.type == "content_block_delta" and chunk.delta.type == "text_delta":  # type: ignore
                yield chunk.delta.text  # type: ignore

    @property
    def text_stream(self) -> Union[Iterator[str], AsyncIterator[str]]:
        return self.__stream_text__()


def create_stream_wrapper(
    name: str,
) -> typing.Callable[[typing.Callable], typing.Callable]:
    def wrapper(fn: typing.Callable) -> typing.Callable:
        op = weave.op()(fn)
        op.name = name  # type: ignore
        return add_accumulator(
            op,  # type: ignore
            make_accumulator=lambda _: anthropic_stream_accumulator,
            should_accumulate=lambda _: True,
            iterator_wrapper=AnthropicIteratorWrapper,  # type: ignore
        )

    return wrapper


anthropic_patcher = MultiPatcher(
    [
        # Patch the sync messages.create method for all messages.create methods
        SymbolPatcher(
            lambda: importlib.import_module("anthropic.resources.messages"),
            "Messages.create",
            create_wrapper_sync(name="anthropic.Messages.create"),
        ),
        SymbolPatcher(
            lambda: importlib.import_module("anthropic.resources.messages"),
            "AsyncMessages.create",
            create_wrapper_async(name="anthropic.AsyncMessages.create"),
        ),
        SymbolPatcher(
            lambda: importlib.import_module("anthropic.resources.messages"),
            "Messages.stream",
            create_stream_wrapper(name="anthropic.Messages.stream"),
        ),
        SymbolPatcher(
            lambda: importlib.import_module("anthropic.resources.messages"),
            "AsyncMessages.stream",
            create_stream_wrapper(name="anthropic.AsyncMessages.stream"),
        ),
    ]
)



# Official interface for interacting with the W&B API. All
# Weave interactions with the Weave API should go through this
# module.

import contextlib
import contextvars
import os
import typing

import aiohttp
import gql
import graphql
from gql.transport.aiohttp import AIOHTTPTransport
from gql.transport.requests import RequestsHTTPTransport
from requests.auth import HTTPBasicAuth

from weave.legacy.weave import errors
from weave.legacy.weave import environment as weave_env
from weave.legacy.weave import wandb_client_api, engine_trace

# Importing at the top-level namespace so other files can import from here.
from weave.legacy.weave.context_state import WandbApiContext, _wandb_api_context

tracer = engine_trace.tracer()  # type: ignore


def set_wandb_api_context(
    user_id: typing.Optional[str],
    api_key: typing.Optional[str],
    headers: typing.Optional[dict],
    cookies: typing.Optional[dict],
) -> typing.Optional[contextvars.Token[typing.Optional[WandbApiContext]]]:
    cur_ctx = get_wandb_api_context()
    if cur_ctx:
        # WANDB API context is only allowed to be set once per thread, since we
        # need to use thread local storage to communicate the context to the wandb
        # lib right now.
        return None
    wandb_client_api.set_wandb_thread_local_api_settings(api_key, cookies, headers)
    return _wandb_api_context.set(WandbApiContext(user_id, api_key, headers, cookies))


def reset_wandb_api_context(
    token: typing.Optional[contextvars.Token[typing.Optional[WandbApiContext]]],
) -> None:
    if token is None:
        return
    wandb_client_api.reset_wandb_thread_local_api_settings()
    _wandb_api_context.reset(token)


@contextlib.contextmanager
def wandb_api_context(
    ctx: typing.Optional[WandbApiContext],
) -> typing.Generator[None, None, None]:
    if ctx:
        token = set_wandb_api_context(
            ctx.user_id, ctx.api_key, ctx.headers, ctx.cookies
        )
    try:
        yield
    finally:
        if ctx:
            reset_wandb_api_context(token)


def get_wandb_api_context() -> typing.Optional[WandbApiContext]:
    return _wandb_api_context.get()


def init() -> typing.Optional[contextvars.Token[typing.Optional[WandbApiContext]]]:
    cookie = weave_env.weave_wandb_cookie()
    headers = weave_env.weave_wandb_gql_headers()
    if cookie:
        # This is a special case for testing. It should never be used in production.
        cookies = {"wandb": cookie}
        headers = {
            "use-admin-privileges": "true",
            "x-origin": "https://app.wandb.test",
            **headers,
        }
        return set_wandb_api_context("admin", None, headers, cookies)
    api_key = weave_env.weave_wandb_api_key()
    if api_key:
        return set_wandb_api_context("admin", api_key, None, None)
    return None


@contextlib.contextmanager
def from_environment() -> typing.Generator[None, None, None]:
    token = init()
    try:
        yield
    finally:
        if token:
            reset_wandb_api_context(token)


class WandbApiAsync:
    def __init__(self) -> None:
        self.connector = aiohttp.TCPConnector(limit=50)

    async def query(
        self, query: graphql.DocumentNode, **kwargs: typing.Any
    ) -> typing.Any:
        wandb_context = get_wandb_api_context()
        headers = None
        cookies = None
        auth = None
        if wandb_context is not None:
            headers = wandb_context.headers
            cookies = wandb_context.cookies
            if wandb_context.api_key is not None:
                auth = aiohttp.BasicAuth("api", wandb_context.api_key)
        # TODO: This is currently used by our FastAPI auth helper, there's probably a better way.
        api_key_override = kwargs.pop("api_key", None)
        if api_key_override:
            auth = aiohttp.BasicAuth("api", api_key_override)
        url_base = weave_env.wandb_base_url()
        transport = AIOHTTPTransport(
            url=url_base + "/graphql",
            client_session_args={
                "connector": self.connector,
                "connector_owner": False,
            },
            headers=headers,
            cookies=cookies,
            auth=auth,
        )
        # Warning: we do not use the recommended context manager pattern, because we're
        # using connector_owner to tell the session not to close our connection pool.
        # There is a bug in aiohttp that causes session close to hang for the ssl_close_timeout
        # which is 10 seconds by default. See issue: https://github.com/graphql-python/gql/issues/381
        # Closing the session just closes the connector, which we don't want anyway, so we don't
        # bother.
        client = gql.Client(transport=transport, fetch_schema_from_transport=False)
        session = await client.connect_async(reconnecting=False)  # type: ignore
        result = await session.execute(query, kwargs)
        # Manually reset the connection, bypassing the SSL bug, avoiding ERROR:asyncio:Unclosed client session
        await transport.session.close()
        return result

    SERVER_INFO_QUERY = gql.gql(
        """
        query ServerInfo {
            serverInfo {
            frontendHost
            }
        }
        """
    )

    async def server_info(self) -> typing.Any:
        return await self.query(self.SERVER_INFO_QUERY)

    ARTIFACT_MANIFEST_QUERY = gql.gql(
        """
        query ArtifactManifest(
            $entityName: String!,
            $projectName: String!,
            $name: String!
        ) {
            project(name: $projectName, entityName: $entityName) {
                artifact(name: $name) {
                    currentManifest {
                        id
                        file {
                            directUrl
                        }
                    }
                }
            }
        }
        """
    )

    async def artifact_manifest_url(
        self, entity_name: str, project_name: str, name: str
    ) -> typing.Optional[str]:
        try:
            result = await self.query(
                self.ARTIFACT_MANIFEST_QUERY,
                entityName=entity_name,
                projectName=project_name,
                name=name,
            )
        except gql.transport.exceptions.TransportQueryError as e:
            return None
        project = result["project"]
        if project is None:
            return None
        artifact = project["artifact"]
        if artifact is None:
            return None
        current_manifest = artifact["currentManifest"]
        if current_manifest is None:
            return None
        file = current_manifest["file"]
        if file is None:
            return None
        return file["directUrl"]

    ARTIFACT_MANIFEST_FROM_ID_QUERY = gql.gql(
        """
        query ArtifactManifestFromID(
            $artifactID: ID!
        ) {
            artifact(id: $artifactID) {
                currentManifest {
                    id
                    file {
                        directUrl
                    }
                }
            }
        }
        """
    )

    async def artifact_manifest_url_from_id(self, art_id: str) -> typing.Optional[str]:
        try:
            result = await self.query(
                self.ARTIFACT_MANIFEST_FROM_ID_QUERY, artifactID=art_id
            )
        except gql.transport.exceptions.TransportQueryError as e:
            return None
        artifact = result["artifact"]
        if artifact is None:
            return None
        current_manifest = artifact["currentManifest"]
        if current_manifest is None:
            return None
        file = current_manifest["file"]
        if file is None:
            return None
        return file["directUrl"]

    VIEWER_DEFAULT_ENTITY_QUERY = gql.gql(
        """
        query DefaultEntity {
            viewer {
                defaultEntity {
                    name
                }
            }
        }
        """
    )

    async def default_entity_name(self) -> typing.Optional[str]:
        try:
            result = await self.query(self.VIEWER_DEFAULT_ENTITY_QUERY)
        except gql.transport.exceptions.TransportQueryError as e:
            return None
        try:
            return result.get("viewer", {}).get("defaultEntity", {}).get("name", None)
        except AttributeError:
            return None

    ENTITY_ACCESS_QUERY = gql.gql(
        """
        query Entity($entityName: String!) {
            viewer { username }
            entity(name: $entityName) { readOnly }
        }
        """
    )

    async def can_access_entity(
        self, entity: str, api_key: typing.Optional[str]
    ) -> bool:
        try:
            result = await self.query(
                self.ENTITY_ACCESS_QUERY, entityName=entity, api_key=api_key
            )
        except gql.transport.exceptions.TransportQueryError as e:
            return False
        return (
            result.get("viewer")
            and result.get("entity", {}).get("readOnly", True) == False
        )


class WandbApi:
    def query(self, query: graphql.DocumentNode, **kwargs: typing.Any) -> typing.Any:
        wandb_context = get_wandb_api_context()
        headers = None
        cookies = None
        auth = None
        if wandb_context is not None:
            headers = wandb_context.headers
            cookies = wandb_context.cookies
            if wandb_context.api_key is not None:
                auth = HTTPBasicAuth("api", wandb_context.api_key)
        url_base = weave_env.wandb_base_url()
        transport = RequestsHTTPTransport(
            url=url_base + "/graphql", headers=headers, cookies=cookies, auth=auth
        )
        # Warning: we do not use the recommended context manager pattern, because we're
        # using connector_owner to tell the session not to close our connection pool.
        # There is a bug in aiohttp that causes session close to hang for the ssl_close_timeout
        # which is 10 seconds by default. See issue: https://github.com/graphql-python/gql/issues/381
        # Closing the session just closes the connector, which we don't want anyway, so we don't
        # bother.
        client = gql.Client(transport=transport, fetch_schema_from_transport=False)
        session = client.connect_sync()  # type: ignore
        return session.execute(query, kwargs)

    SERVER_INFO_QUERY = gql.gql(
        """
        query ServerInfo {
            serverInfo {
            frontendHost
            }
        }
        """
    )

    def server_info(self) -> typing.Any:
        return self.query(self.SERVER_INFO_QUERY)

    ARTIFACT_MANIFEST_QUERY = gql.gql(
        """
        query ArtifactManifest(
            $entityName: String!,
            $projectName: String!,
            $name: String!
        ) {
            project(name: $projectName, entityName: $entityName) {
                artifact(name: $name) {
                    currentManifest {
                        id
                        file {
                            directUrl
                        }
                    }
                }
            }
        }
        """
    )

    def artifact_manifest_url(
        self, entity_name: str, project_name: str, name: str
    ) -> typing.Optional[str]:
        try:
            result = self.query(
                self.ARTIFACT_MANIFEST_QUERY,
                entityName=entity_name,
                projectName=project_name,
                name=name,
            )
        except gql.transport.exceptions.TransportQueryError as e:
            return None
        project = result["project"]
        if project is None:
            return None
        artifact = project["artifact"]
        if artifact is None:
            return None
        current_manifest = artifact["currentManifest"]
        if current_manifest is None:
            return None
        file = current_manifest["file"]
        if file is None:
            return None
        return file["directUrl"]

    ARTIFACT_MANIFEST_FROM_ID_QUERY = gql.gql(
        """
        query ArtifactManifestFromID(
            $artifactID: ID!
        ) {
            artifact(id: $artifactID) {
                currentManifest {
                    id
                    file {
                        directUrl
                    }
                }
            }
        }
        """
    )

    def artifact_manifest_url_from_id(self, art_id: str) -> typing.Optional[str]:
        try:
            result = self.query(self.ARTIFACT_MANIFEST_FROM_ID_QUERY, artifactID=art_id)
        except gql.transport.exceptions.TransportQueryError as e:
            return None
        artifact = result["artifact"]
        if artifact is None:
            return None
        current_manifest = artifact["currentManifest"]
        if current_manifest is None:
            return None
        file = current_manifest["file"]
        if file is None:
            return None
        return file["directUrl"]

    VIEWER_DEFAULT_ENTITY_QUERY = gql.gql(
        """
        query DefaultEntity {
            viewer {
                username
                defaultEntity {
                    name
                }
            }
        }
        """
    )

    def default_entity_name(self) -> typing.Optional[str]:
        try:
            result = self.query(self.VIEWER_DEFAULT_ENTITY_QUERY)
        except gql.transport.exceptions.TransportQueryError as e:
            return None
        try:
            return result.get("viewer", {}).get("defaultEntity", {}).get("name", None)
        except AttributeError:
            return None

    def username(self) -> typing.Optional[str]:
        try:
            result = self.query(self.VIEWER_DEFAULT_ENTITY_QUERY)
        except gql.transport.exceptions.TransportQueryError as e:
            return None

        return result.get("viewer", {}).get("username", None)


async def get_wandb_api() -> WandbApiAsync:
    return WandbApiAsync()


def get_wandb_api_sync() -> WandbApi:
    return WandbApi()



import base64
import cProfile
import logging
import os
import pathlib
import sys
import time
import traceback
import typing
import urllib.parse
import zlib

import requests
import wandb
from flask import (
    Blueprint,
    Flask,
    Response,
    abort,
    json,
    redirect,
    request,
    send_from_directory,
)
from flask_cors import CORS
from werkzeug.exceptions import HTTPException

from weave.legacy.weave import (
    context_state,
    engine_trace,
    environment,
    errors,
    filesystem,
    graph,
    logs,
    registry_mem,
    server,
    storage,
    util,
    value_or_error,
    wandb_api,
    weavejs_fixes,
)
from weave.legacy.weave.language_features.tagging import tag_store
from weave.legacy.weave.server_error_handling import (
    client_safe_http_exceptions_as_werkzeug,
)

logger = logging.getLogger(__name__)

WEAVE_CLIENT_CACHE_KEY_HEADER = "x-weave-client-cache-key"

# PROFILE_DIR = "/tmp/weave/profile"
PROFILE_DIR = None
if PROFILE_DIR is not None:
    pathlib.Path(PROFILE_DIR).mkdir(parents=True, exist_ok=True)

ERROR_STR_LIMIT = 10000

tracer = engine_trace.tracer()


def custom_dd_patch():
    import wandb_gql  # type: ignore[import]
    import wandb_graphql  # type: ignore[import]

    orig_execute = wandb_gql.Client.execute

    def execute(self, document, *args, **kwargs):
        with tracer.trace(
            "wandb_gql.Client.execute",
        ) as span:
            span.set_tag("document", wandb_graphql.print_ast(document))
            span.set_tag("variable_values", kwargs.get("variable_values", {}))
            return orig_execute(self, document, *args, **kwargs)

    wandb_gql.Client.execute = execute


if engine_trace.datadog_is_enabled():
    # Don't import this if you don't have an Agent! You'll get weird
    # crashes
    import ddtrace

    # patch all libraries to datadog except logging
    ddtrace.patch_all(logging=False)

    if not environment.disable_weave_pii():
        # patch logging to datadog, if not disabled
        ddtrace.patch(logging=True)

    custom_dd_patch()


# Ensure these are imported and registered
from weave.legacy.weave import ops

# NOTE: Fixes flask dev server's auto-reload capability, by forcing it to use
# stat mode instead of watchdog mode. It turns out that "import wandb" breaks
# users of watchdog somehow. We'll need to fix that in wandb.
# A wandb fix should go out in 0.13.5.
# TODO: remove this hack when wandb 0.13.5 is released.
if os.environ.get("FLASK_DEBUG"):
    print(
        "!!! Weave server removing watchdog from sys.path for development mode. This could break other libraries"
    )
    sys.modules["watchdog.observers"] = None  # type: ignore


static_folder = os.path.join(os.path.dirname(__file__), "frontend")
blueprint = Blueprint("weave", "weave-server", static_folder=static_folder)


def import_ecosystem():
    from weave.legacy.weave import ops, panels, panels_py

    # Attempt to import MVP ecosystem modules
    try:
        from weave.legacy.weave.ecosystem import langchain, replicate
    except ImportError:
        pass

    if not util.parse_boolean_env_var("WEAVE_SERVER_DISABLE_ECOSYSTEM"):
        # try:
        #     from weave.weaveflow import faiss

        #     # Turn eager mode back off.
        #     context_state._eager_mode.set(False)
        # except (ImportError, OSError, wandb.Error):
        #     print("Error: Couldn't import faiss module for Weaveflow.")
        try:
            from weave.legacy.weave.ecosystem import all
        except (ImportError, OSError, wandb.Error):
            pass


def log_system_info():
    WEAVE_SERVER_URL = os.environ.get("WEAVE_SERVER_URL")
    WANDB_BASE_URL = os.environ.get("WANDB_BASE_URL", "http://api.wandb.ai")
    WEAVE_LOCAL_ARTIFACT_DIR = os.environ.get("WEAVE_LOCAL_ARTIFACT_DIR")
    dot_netrc_exists = os.path.exists(os.path.expanduser("~/.netrc"))
    underscore_netrc_exists = os.path.exists(os.path.expanduser("~/_netrc"))
    WANDB_API_KEY = "REDACTED" if os.environ.get("WANDB_API_KEY") else None
    WEAVE_WANDB_COOKIE = "REDACTED" if os.environ.get("WEAVE_WANDB_COOKIE") else None
    WEAVE_WANDB_GQL_HEADERS = environment.weave_wandb_gql_headers()
    FRONTEND_ENV = json.dumps(frontend_env(), indent=2)

    logger.info("Network Config:")
    logger.info(f"  WEAVE_SERVER_URL        = {WEAVE_SERVER_URL}")
    logger.info(f"  WANDB_BASE_URL          = {WANDB_BASE_URL}")
    logger.info(f"  WEAVE_WANDB_GQL_HEADERS = {WEAVE_WANDB_GQL_HEADERS}")

    logger.info("Cache Config:")
    logger.info(f"  WEAVE_LOCAL_ARTIFACT_DIR  = {WEAVE_LOCAL_ARTIFACT_DIR}")

    logger.info("Auth Config:")
    logger.info(f"  ~/.netrc exists     = {dot_netrc_exists}")
    logger.info(f"  ~/_netrc exists     = {underscore_netrc_exists}")
    logger.info(f"  WANDB_API_KEY       = {WANDB_API_KEY}")
    logger.info(f"  WEAVE_WANDB_COOKIE  = {WEAVE_WANDB_COOKIE}")
    logger.info(f"  FRONTEND_ENV        = {FRONTEND_ENV}")


def make_app():
    logs.configure_logger()
    import_ecosystem()

    logger.info("Starting weave server")
    log_system_info()

    app = Flask(__name__)
    app.register_blueprint(blueprint)
    # Very important! We rely on key ordering on both sides!
    # Flask < 2.2 doesn't have app.json
    if hasattr(app, "json"):
        app.json.sort_keys = False
    else:
        app.config["JSON_SORT_KEYS"] = False

    CORS(app, supports_credentials=True)

    return app


class OpsCache(typing.TypedDict):
    updated_at: float
    ops_data: typing.Optional[dict]


ops_cache: typing.Optional[OpsCache] = None


@blueprint.route("/__weave/ops", methods=["GET"])
def list_ops():
    global ops_cache
    with wandb_api.from_environment():
        # TODO: this is super slow.
        if not environment.wandb_production():
            registry_mem.memory_registry.load_saved_ops()
            pass
        if (
            ops_cache is None
            or ops_cache["updated_at"] < registry_mem.memory_registry.updated_at()
        ):
            ops = registry_mem.memory_registry.list_ops()
            ret = []
            for op in ops:
                try:
                    serialized_op = op.to_dict()
                except errors.WeaveSerializeError:
                    continue
                ret.append(serialized_op)
            ops_cache = {
                "updated_at": registry_mem.memory_registry.updated_at(),
                "data": ret,
            }
    return ops_cache


class ErrorDetailsDict(typing.TypedDict):
    message: str
    error_type: str
    traceback: list[str]
    # sentry_id: typing.Union[str, None]


class ResponseDict(typing.TypedDict):
    data: list[typing.Any]
    errors: list[ErrorDetailsDict]
    node_to_error: dict[int, int]


def _exception_to_error_details(
    e: Exception, sentry_id: typing.Union[str, None]
) -> ErrorDetailsDict:
    return {
        "error_type": type(e).__name__,
        "message": str(e),
        "traceback": traceback.format_tb(e.__traceback__),
        # "sentry_id": sentry_id,
    }


def _value_or_errors_to_response(
    vore: value_or_error.ValueOrErrors,
) -> ResponseDict:
    error_lookup: dict[Exception, typing.Tuple[int, typing.Union[str, None]]] = {}
    node_errors: dict[int, int] = {}
    data: list[typing.Any] = []
    for val, error in vore.iter_items():
        if error != None:
            error = typing.cast(Exception, error)
            data.append(None)
            if error in error_lookup:
                error_ndx = error_lookup[error][0]
            else:
                sentry_id = util.capture_exception_with_sentry_if_available(error, ())
                error_ndx = len(error_lookup)
                error_lookup[error] = (error_ndx, sentry_id)
            node_errors[len(data) - 1] = error_ndx
        else:
            data.append(val)
    return {
        "data": data,
        "errors": [
            _exception_to_error_details(k, e_sentry_id)
            for k, (e_ndx, e_sentry_id) in error_lookup.items()
        ],
        "node_to_error": node_errors,
    }


def _log_errors(
    processed_response: ResponseDict, nodes: value_or_error.ValueOrErrors[graph.Node]
):
    errors: list[dict] = []

    for error in processed_response["errors"]:
        errors.append(
            {
                "message": error["message"],
                "error_type": error["error_type"],
                "traceback": error["traceback"],
                "error_tag": "node_execution_error",
                "node_strs": [],
                # "sentry_id": error["sentry_id"],
            }
        )

    for node_ndx, error_ndx in processed_response["node_to_error"].items():
        try:
            node_str = graph.node_expr_str(graph.map_const_nodes_to_x(nodes[node_ndx]))
            errors[error_ndx]["node_strs"].append(node_str)
        except Exception:
            pass

    for error_dict in errors:
        # This should be logged to DD, but 1 log per error
        # class, not 1 log per error.
        logging.error(error_dict)


def _get_client_cache_key_from_request(request):
    # Uncomment to set default to 15 second cache duration
    client_cache_key = None  # str(int(time.time() // 15))
    if WEAVE_CLIENT_CACHE_KEY_HEADER in request.headers:
        client_cache_key = request.headers[WEAVE_CLIENT_CACHE_KEY_HEADER]
    return client_cache_key


@blueprint.route("/__weave/execute", methods=["POST"])
def execute():
    """Execute endpoint used by WeaveJS."""
    with tracer.trace("read_request"):
        req_bytes = request.data
    req_compressed = zlib.compress(req_bytes)
    req_b64 = base64.b64encode(req_compressed).decode("ascii")
    logging.info(
        "Execute request (zlib): %s",
        req_b64,
    )

    if not request.json:
        abort(400, "Request body must be JSON.")
    if "graphs" not in request.json:
        abort(400, "Request body must contain a 'graphs' key.")

    # Simulate browser/server latency
    # import time
    # time.sleep(0.1)

    # use a single memartifact to serialize the entire response.
    # fixes https://weights-biases.sentry.io/issues/4022569419

    execute_args = {
        "request": request.json,
        "deref": True,
        "serialize_fn": storage.make_js_serializer(),
    }
    root_span = tracer.current_root_span()
    tag_store.record_current_tag_store_size()

    client_cache_key = _get_client_cache_key_from_request(request)

    if not PROFILE_DIR:
        start_time = time.time()
        with client_safe_http_exceptions_as_werkzeug():
            with context_state.set_client_cache_key(client_cache_key):
                response = server.handle_request(**execute_args)
        elapsed = time.time() - start_time
    else:
        # Profile the request and add a link to local snakeviz to the trace.
        profile = cProfile.Profile()
        start_time = time.time()
        try:
            with client_safe_http_exceptions_as_werkzeug():
                with context_state.set_client_cache_key(client_cache_key):
                    response = profile.runcall(server.handle_request, **execute_args)
        finally:
            elapsed = time.time() - start_time
            profile_filename = f"/tmp/weave/profile/execute.{start_time*1000:.0f}.{elapsed*1000:.0f}ms.prof"
            profile.dump_stats(profile_filename)
            if root_span:
                root_span.set_tag(
                    "profile_url",
                    "http://localhost:8080/snakeviz/"
                    + urllib.parse.quote(profile_filename),
                )

    fixed_response = response.results.safe_map(weavejs_fixes.fixup_data)
    response_payload = _value_or_errors_to_response(fixed_response)

    if root_span is not None:
        root_span.set_metric("request_size", len(req_bytes), True)
        root_span.set_metric("node_count", len(response_payload["data"]), True)
        root_span.set_metric(
            "error_count",
            len(response_payload["node_to_error"]),
            True,
        )

    _log_errors(response_payload, response.nodes)

    if request.headers.get("x-weave-include-execution-time"):
        response_payload["execution_time"] = (elapsed) * 1000

    return response_payload


@blueprint.route("/__weave/execute/v2", methods=["POST"])
def execute_v2():
    """Execute endpoint used by Weave Python"""
    # print('REQUEST', request, request.json)
    if not request.json or "graphs" not in request.json:
        abort(400)
    response = server.handle_request(request.json, deref=True)
    # print("RESPONSE BEFORE SERI", response)

    return {"data": response.results.unwrap()}


@blueprint.route("/__weave/file/<path:path>")
def send_local_file(path):
    try:
        # Retrieve and normalize the local artifacts path
        local_artifacts_path = pathlib.Path(filesystem.get_filesystem_dir()).resolve(
            strict=True
        )

        # Construct the full absolute path of the requested file
        requested_path = (local_artifacts_path / path).resolve(strict=True)

        # Ensure the requested path is within the local artifacts directory
        if not str(requested_path).startswith(str(local_artifacts_path)):
            abort(403)

        # Send the file from the directory
        return send_from_directory(
            local_artifacts_path, str(requested_path.relative_to(local_artifacts_path))
        )
    except (errors.WeaveAccessDeniedError, FileNotFoundError):
        abort(403)


@blueprint.before_request
def _disable_eager_mode():
    context_state._eager_mode.set(False)


def frontend_env():
    """If you add vars here, make sure to define their types in weave-js/src/config.ts"""
    return {
        "PREFIX": environment.weave_link_prefix(),
        "ANALYTICS_DISABLED": environment.analytics_disabled(),
        "ONPREM": environment.weave_onprem(),
        "WEAVE_BACKEND_HOST": environment.weave_backend_host(),
        "TRACE_BACKEND_BASE_URL": environment.trace_backend_base_url(),
        "WANDB_BASE_URL": environment.wandb_frontend_base_url(),
        "DD_ENV": environment.dd_env(),
        "ENV_IS_CI": environment.env_is_ci(),
    }


def legacy_weave_redirect():
    return redirect("https://wandb.github.io/weave/", code=302)


@blueprint.route("/__frontend", defaults={"path": None})
@blueprint.route("/__frontend/<path:path>")
def frontend(path):
    # Redirect to the new docs site
    if not environment.env_is_ci():
        return legacy_weave_redirect()

    """Serve the frontend with a simple fileserver over HTTP."""
    # We serve up a dynamic env.js file before all other js.
    if path is not None and path.endswith("env.js"):
        js = f"window.WEAVE_CONFIG = {json.dumps(frontend_env())}"
        return Response(js, mimetype="application/javascript")
    full_path = pathlib.Path(blueprint.static_folder) / path
    # prevent path traversal
    if not full_path.resolve().is_relative_to(blueprint.static_folder):
        return abort(403)
    if path and full_path.exists():
        return send_from_directory(blueprint.static_folder, path)
    else:
        return send_from_directory(blueprint.static_folder, "index.html")


@blueprint.route("/", defaults={"path": None})
@blueprint.route("/<path:path>")
def root_frontend(path):
    # Redirect to the new docs site
    if not environment.env_is_ci():
        return legacy_weave_redirect()

    if request.args.get("unsetBetaVersion") is not None:
        resp = redirect_without_query_param("unsetBetaVersion")
        resp.set_cookie("betaVersion", "", max_age=0)
        return resp

    new_beta_version = request.args.get("betaVersion")
    if new_beta_version is not None:
        resp = redirect_without_query_param("betaVersion")
        resp.set_cookie("betaVersion", new_beta_version)
        return resp

    # To support server cases where we're mounted under an existing path, i.e.
    # /wandb/weave, then index.html will load something like /wandb/weave/.../env.js
    if path is not None:
        path = path.split("/")[-1]
    if path == "env.js":
        js = f"window.WEAVE_CONFIG = {json.dumps(frontend_env())}"
        return Response(js, mimetype="application/javascript")

    beta_version = request.cookies.get("betaVersion")
    if beta_version is not None:
        beta_version = beta_version[:9]
        content = requests.get(
            f"https://cdn.wandb.ai/weave/{beta_version}/index.html",
            stream=True,
        ).content
        return Response(content, mimetype="text/html")

    return send_from_directory(blueprint.static_folder, "index.html")


def redirect_without_query_param(param: str):
    qs_pairs = []
    for k, v in request.args.items():
        if k != param:
            qs_pairs.append(f"{k}={v}")
    qs = "&".join(qs_pairs)
    resp = redirect(f"{request.path}?{qs}")
    return resp


@blueprint.route("/__weave/hello")
def hello():
    return "hello"


@blueprint.route("/__weave/wb_viewer", methods=["POST"])
def wb_viewer():
    wandb_api.init()
    current_context = wandb_api.get_wandb_api_context()
    if not current_context:
        with wandb_api.from_environment():
            current_context = wandb_api.get_wandb_api_context()
    authenticated = current_context is not None
    user_id = None if current_context is None else current_context.user_id
    return {"authenticated": authenticated, "user_id": user_id}


app = make_app()

if os.getenv("WEAVE_SERVER_DEBUG"):

    @app.errorhandler(HTTPException)
    def handle_exception(e):
        """Return JSON instead of HTML for HTTP errors."""
        # start with the correct headers and status code from the error
        response = e.get_response()
        # replace the body with JSON
        response.data = json.dumps(
            {
                "code": e.code,
                "name": e.name,
                "description": e.description,
                "exc_info": traceback.format_exc(),
            }
        )
        response.content_type = "application/json"
        return response


if __name__ == "__main__":
    app.run(port=9994)



import datetime
import typing

import pyarrow as pa
import pyarrow.compute as pc

from weave.legacy.weave import weave_types as types
from weave.legacy.weave.api import op
from weave.legacy.weave import timestamp as weave_timestamp
from weave.legacy.weave.arrow.list_ import ArrowWeaveList, ArrowWeaveListType
from weave.legacy.weave.decorator_arrow_op import arrow_op
from weave.legacy.weave.ops_arrow import util

ARROW_WEAVE_LIST_NUMBER_TYPE = ArrowWeaveListType(types.Number())
ARROW_WEAVE_LIST_BOOLEAN_TYPE = ArrowWeaveListType(types.Boolean())
ARROW_WEAVE_LIST_STRING_TYPE = ArrowWeaveListType(types.String())

unary_input_type = {
    "self": ARROW_WEAVE_LIST_NUMBER_TYPE,
}
binary_input_type = {
    "self": ARROW_WEAVE_LIST_NUMBER_TYPE,
    "other": types.UnionType(types.Number(), ARROW_WEAVE_LIST_NUMBER_TYPE),
}

null_consuming_binary_input_type = {
    "self": ArrowWeaveListType(types.optional(types.Number())),
    "other": types.UnionType(
        types.optional(types.Number()),
        ArrowWeaveListType(types.optional(types.Number())),
    ),
}

null_consuming_binary_input_type_right = {
    "self": types.optional(types.Number()),
    "other": ArrowWeaveListType(types.optional(types.Number())),
}


def self_type_output_type_fn(input_types):
    return input_types["self"]


# TODO: weirdly need to name this "<something>-add" since that's what the base
# Number op does. But we'd like to get rid of that requirement so we don't
# need name= for any ops!
@arrow_op(
    name="ArrowWeaveListNumber-add",
    input_type=binary_input_type,
    output_type=self_type_output_type_fn,
)
def __add__(self, other):
    if isinstance(other, ArrowWeaveList):
        other = other._arrow_data
    return ArrowWeaveList(
        pc.add(self._arrow_data, other),
        types.NoneType() if other == None else types.Number(),
        self._artifact,
    )


@op(
    name="ArrowWeaveListNumber_right-add",
    output_type=lambda input_type: input_type["right"],
)
def number_right_add(
    left: typing.Optional[typing.Union[int, float]],
    right: ArrowWeaveList[typing.Optional[typing.Union[int, float]]],
):
    return ArrowWeaveList(
        pc.add(left, right._arrow_data_asarray_no_tags()),
        types.Number(),
        right._artifact,
    )


@arrow_op(
    name="ArrowWeaveListNumber-mult",
    input_type=binary_input_type,
    output_type=self_type_output_type_fn,
)
def __mul__(self, other):
    if isinstance(other, ArrowWeaveList):
        other = other._arrow_data
    return ArrowWeaveList(
        pc.multiply(self._arrow_data, other),
        types.NoneType() if other == None else types.Number(),
        self._artifact,
    )


@op(
    name="ArrowWeaveListNumber_right-mult",
    output_type=lambda input_type: input_type["right"],
)
def number_right_mult(
    left: typing.Optional[typing.Union[int, float]],
    right: ArrowWeaveList[typing.Optional[typing.Union[int, float]]],
):
    return ArrowWeaveList(
        pc.multiply(left, right._arrow_data_asarray_no_tags()),
        types.Number(),
        right._artifact,
    )


@arrow_op(
    name="ArrowWeaveListNumber-div",
    input_type=binary_input_type,
    output_type=ArrowWeaveListType(types.Float()),
)
def __truediv__(self, other):
    if isinstance(other, ArrowWeaveList):
        other = other._arrow_data.cast(pa.float64())
    return ArrowWeaveList(
        pc.divide(self._arrow_data.cast(pa.float64()), other),
        types.Float(),
        self._artifact,
    )


@arrow_op(
    name="ArrowWeaveListNumber-sub",
    input_type=binary_input_type,
    output_type=self_type_output_type_fn,
)
def __sub__(self, other):
    if isinstance(other, ArrowWeaveList):
        other = other._arrow_data
    return ArrowWeaveList(
        pc.subtract(self._arrow_data, other),
        types.NoneType() if other == None else types.Number(),
        self._artifact,
    )


@op(
    name="ArrowWeaveListNumber_right-sub",
    output_type=lambda input_type: input_type["right"],
)
def number_right_sub(
    left: typing.Optional[typing.Union[int, float]],
    right: ArrowWeaveList[typing.Optional[typing.Union[int, float]]],
):
    return ArrowWeaveList(
        pc.subtract(left, right._arrow_data_asarray_no_tags()),
        types.Number(),
        right._artifact,
    )


@arrow_op(
    name="ArrowWeaveListNumber-powBinary",
    input_type=binary_input_type,
    output_type=self_type_output_type_fn,
)
def __pow__(self, other):
    if isinstance(other, ArrowWeaveList):
        other = other._arrow_data
    return ArrowWeaveList(
        pc.power(self._arrow_data, other), types.Number(), self._artifact
    )


@arrow_op(
    name="ArrowWeaveListNumber-notEqual",
    input_type=null_consuming_binary_input_type,
    output_type=ARROW_WEAVE_LIST_BOOLEAN_TYPE,
)
def __ne__(self, other):
    if isinstance(other, ArrowWeaveList):
        other = other._arrow_data
    result = util.not_equal(self._arrow_data, other)
    return ArrowWeaveList(result, types.Boolean(), self._artifact)


@arrow_op(
    name="ArrowWeaveListNumber-equal",
    input_type=null_consuming_binary_input_type,
    output_type=ARROW_WEAVE_LIST_BOOLEAN_TYPE,
)
def __eq__(self, other):
    if isinstance(other, ArrowWeaveList):
        other = other._arrow_data
    result = util.equal(self._arrow_data, other)
    return ArrowWeaveList(result, types.Boolean(), self._artifact)


@arrow_op(
    name="ArrowWeaveListNumber-greater",
    input_type=binary_input_type,
    output_type=ARROW_WEAVE_LIST_BOOLEAN_TYPE,
)
def __gt__(self, other):
    if isinstance(other, ArrowWeaveList):
        other = other._arrow_data
    return ArrowWeaveList(
        pc.greater(self._arrow_data, other), types.Boolean(), self._artifact
    )


@arrow_op(
    name="ArrowWeaveListNumber-greaterEqual",
    input_type=binary_input_type,
    output_type=ARROW_WEAVE_LIST_BOOLEAN_TYPE,
)
def __ge__(self, other):
    if isinstance(other, ArrowWeaveList):
        other = other._arrow_data
    return ArrowWeaveList(
        pc.greater_equal(self._arrow_data, other), types.Boolean(), self._artifact
    )


@arrow_op(
    name="ArrowWeaveListNumber-less",
    input_type=binary_input_type,
    output_type=ARROW_WEAVE_LIST_BOOLEAN_TYPE,
)
def __lt__(self, other):
    if isinstance(other, ArrowWeaveList):
        other = other._arrow_data
    return ArrowWeaveList(
        pc.less(self._arrow_data, other), types.Boolean(), self._artifact
    )


@arrow_op(
    name="ArrowWeaveListNumber-lessEqual",
    input_type=binary_input_type,
    output_type=ARROW_WEAVE_LIST_BOOLEAN_TYPE,
)
def __le__(self, other):
    if isinstance(other, ArrowWeaveList):
        other = other._arrow_data
    return ArrowWeaveList(
        pc.less_equal(self._arrow_data, other), types.Boolean(), self._artifact
    )


@arrow_op(
    name="ArrowWeaveListNumber-negate",
    input_type=unary_input_type,
    output_type=self_type_output_type_fn,
)
def __neg__(self):
    return ArrowWeaveList(pc.negate(self._arrow_data), types.Number(), self._artifact)


# todo: fix op decorator to not require name here,
# these names are needed because the base class also has an op called floor.
# also true for ceil below
@arrow_op(
    name="ArrowWeaveListNumber-floor",
    input_type=unary_input_type,
    output_type=self_type_output_type_fn,
)
def floor(self):
    return ArrowWeaveList(pc.floor(self._arrow_data), types.Number(), self._artifact)


@arrow_op(
    name="ArrowWeaveListNumber-ceil",
    input_type=unary_input_type,
    output_type=self_type_output_type_fn,
)
def ceil(self):
    return ArrowWeaveList(pc.ceil(self._arrow_data), types.Number(), self._artifact)


@arrow_op(
    name="ArrowWeaveListNumber-round",
    input_type=unary_input_type,
    output_type=self_type_output_type_fn,
)
def round(self):
    return ArrowWeaveList(pc.round(self._arrow_data), types.Number(), self._artifact)


@arrow_op(
    name="ArrowWeaveListNumber-abs",
    input_type={"self": ArrowWeaveListType(types.optional(types.Number()))},
    output_type=self_type_output_type_fn,
)
def abs(self):
    return ArrowWeaveList(pc.abs(self._arrow_data), types.Number(), self._artifact)


@arrow_op(
    name="ArrowWeaveListNumber-toFixed",
    input_type={
        "self": ArrowWeaveListType(types.optional(types.Number())),
        "digits": types.Number(),
    },
    output_type=self_type_output_type_fn,
)
def to_fixed(self, digits):
    return ArrowWeaveList(
        pc.round(self._arrow_data, int(digits)), types.Number(), self._artifact
    )


@op(
    name="ArrowWeaveListNumber-avg",
    input_type={"self": ArrowWeaveListType(types.optional(types.Number()))},
    output_type=types.Number(),
)
def avg(self):
    array = self._arrow_data_asarray_no_tags()
    return pc.mean(array).as_py()


@op(
    name="ArrowWeaveListNumber-sum",
    input_type={"self": ArrowWeaveListType(types.optional(types.Number()))},
    output_type=types.Number(),
)
def sum(self):
    array = self._arrow_data_asarray_no_tags()
    return pc.sum(array).as_py()


@op(
    name="ArrowWeaveListNumber-max",
    input_type={"self": ArrowWeaveListType(types.optional(types.Number()))},
    output_type=types.Number(),
)
def max(self):
    array = self._arrow_data_asarray_no_tags()
    return pc.max(array).as_py()


@op(
    name="ArrowWeaveListNumber-min",
    input_type={"self": ArrowWeaveListType(types.optional(types.Number()))},
    output_type=types.Number(),
)
def min(self):
    array = self._arrow_data_asarray_no_tags()
    return pc.min(array).as_py()


# TODO: Do we need to vectorize these in 1 more dimension?
@op(
    name="ArrowWeaveListNumbers-argmax",
    input_type={"self": ArrowWeaveListType(types.optional(types.Number()))},
    output_type=types.Number(),
)
def argmax(self):
    array = self._arrow_data_asarray_no_tags()
    return pc.index(array, pc.max(array)).as_py()


@op(
    name="ArrowWeaveListNumbers-argmin",
    input_type={"self": ArrowWeaveListType(types.optional(types.Number()))},
    output_type=types.Number(),
)
def argmin(self):
    array = self._arrow_data_asarray_no_tags()
    return pc.index(array, pc.min(array)).as_py()


@op(
    name="ArrowWeaveListNumbers-stddev",
    input_type={"self": ArrowWeaveListType(types.optional(types.Number()))},
    output_type=types.Number(),
)
def stddev(self):
    array = self._arrow_data_asarray_no_tags()
    return pc.stddev(array).as_py()


@arrow_op(
    name="ArrowWeaveListNumber-toTimestamp",
    input_type=unary_input_type,
    output_type=ArrowWeaveListType(types.Timestamp()),
)
def to_timestamp(self):
    data = self._arrow_data
    try:
        data = data.cast("float64")
    except pa.ArrowInvalid:
        # In the case that our data overflows the bounds of a float64,
        # we just proceed with the original data. If this is the case,
        # then the data is already a high enough precision timestamp
        # and we will need to adjust it anyway.
        pass

    def adjustment_mask():
        return pc.or_(
            pc.less(data, weave_timestamp.PY_DATETIME_MIN_MS),
            pc.greater(data, weave_timestamp.PY_DATETIME_MAX_MS),
        )

    mask = adjustment_mask()

    # In practice this should only ever need to run once or twice.
    while pc.sum(mask).as_py() > 0:
        adjusted_data = pc.divide(data, 1000)
        data = pc.if_else(mask, adjusted_data, data)
        mask = adjustment_mask()

    data_as_timestamp = (
        pc.floor(data).cast("int64").cast(pa.timestamp("ms", tz=datetime.timezone.utc))
    )
    return ArrowWeaveList(
        data_as_timestamp,
        types.Timestamp(),
        self._artifact,
    )


@arrow_op(
    name="ArrowWeaveListNumber-toString",
    input_type=unary_input_type,
    output_type=ARROW_WEAVE_LIST_STRING_TYPE,
)
def to_string(self):
    return ArrowWeaveList(
        pc.cast(self._arrow_data, "string"), types.String(), self._artifact
    )



import contextlib
import contextvars
import typing

# Throw an error if op saving encounters an unknonwn condition.
# The default behavior is to warn.
_strict_op_saving: contextvars.ContextVar[bool] = contextvars.ContextVar(
    "_strict_op_saving", default=False
)


def get_strict_op_saving() -> bool:
    return _strict_op_saving.get()


@contextlib.contextmanager
def strict_op_saving(enabled: bool) -> typing.Generator[bool, None, None]:
    token = _strict_op_saving.set(enabled)
    yield _strict_op_saving.get()
    _strict_op_saving.reset(token)



import typing

from weave.legacy.weave import codify
from weave.legacy.weave import weave_types as types
from weave.legacy.weave.api import op, weave_class


# This matches the output type logic of the frontend
# We know that this is going to need refinement, but are
# rolling with it in the first iteration. We need to get to the point
# where we don't have all these conditionals.
def getattr_output_type(input_type):
    self_arg = input_type["self"]

    if isinstance(self_arg, types.Const):
        self_arg = self_arg.val_type

    if isinstance(self_arg, types.ObjectType):
        if not isinstance(input_type["name"], types.Const):
            return types.UnknownType()
        key = input_type["name"].val
        property_types = self_arg.property_types()
        return property_types.get(key, types.NoneType())

    if isinstance(self_arg, types.Any):
        return types.Any()

    # TODO: In particular, this branch should go away since
    # we anticipate getattr will eventually just target the
    # object type. This is a temporary special casing until
    # the type hierarchy is properly worked out. For example,
    # here we special case `property_types` and `members` for
    # the typedDict and union types respectively. Then just
    # blindly return a type otherwise. This is just circumstancially
    # correct for now.
    if isinstance(self_arg, types.Type):
        if isinstance(input_type["name"], types.Const):
            if input_type["name"].val == "property_types":
                return types.Dict(types.String(), types.TypeType())
            elif input_type["name"].val == "members":
                return types.List(types.TypeType())
        return types.TypeType()

    return types.Invalid()


def obj_settattr(self, attr, v):
    setattr(self, attr, v)
    return self


@op(
    name="Object-__getattr__",
    setter=obj_settattr,
    input_type={"self": types.ObjectType()},
    output_type=getattr_output_type,
)
def obj_getattr(self, name: str):
    return getattr(self, name, None)


@op(
    name="__internal__-generateCodeForObject",
    input_type={"obj": types.Any()},
    hidden=True,
)
def generate_code_for_object(obj) -> str:
    return codify.object_to_code(obj)



# A thread that manages all Weave IO (we haven't migrated everything here yet).
# You can perform IO by using the get_sync_client or get_async_client interfaces.

# Warning: if you see errors coming from here, logging in the server process
# is not currently working. You can change logger.error to print to get stack
# traces in local development.
# TODO: Fix

import asyncio
import atexit
import contextlib
import dataclasses
import logging
import multiprocessing
import queue
import threading
import time
import traceback
import typing
import uuid
from typing import Any, Callable, Dict, Iterator, TypeVar

import aioprocessing

from weave.legacy.weave import (
    weave_http,
    filesystem,
    errors,
    engine_trace,
    server_error_handling,
    artifact_wandb,
    async_queue,
    cache,
    context_state,
    uris,
    wandb_api,
    wandb_file_manager,
)

tracer = engine_trace.tracer()  # type: ignore
statsd = engine_trace.statsd()  # type: ignore


QueueItemType = TypeVar("QueueItemType")


class ArtifactMetadata(typing.TypedDict):
    created_at: str


# ServerRequestContext holds the context for server requests
@dataclasses.dataclass
class ServerRequestContext:
    trace_context: typing.Optional[engine_trace.TraceContext]
    wandb_api_context: typing.Optional[wandb_api.WandbApiContext]
    cache_prefix_context: typing.Optional[str]

    @classmethod
    def from_json(cls, json: typing.Any) -> "ServerRequestContext":
        trace_context = engine_trace.new_trace_context()
        if trace_context:
            trace_context.__setstate__(json["trace_context"])
        wandb_api_context = None
        wandb_api_context_json = json.get("wandb_api_context")
        if wandb_api_context_json:
            wandb_api_context = wandb_api.WandbApiContext.from_json(
                wandb_api_context_json
            )
        cache_prefix_context = json.get("cache_prefix_context", None)
        if cache_prefix_context:
            cache_prefix_context = str(cache_prefix_context)
        return cls(
            trace_context=trace_context,
            wandb_api_context=wandb_api_context,
            cache_prefix_context=cache_prefix_context,
        )

    def to_json(self) -> typing.Any:
        trace_context = None
        if self.trace_context:
            trace_context = self.trace_context.__getstate__()
        wandb_ctx = None
        if self.wandb_api_context:
            wandb_ctx = self.wandb_api_context.to_json()
        if self.cache_prefix_context:
            cache_prefix_context = self.cache_prefix_context
        return {
            "trace_context": trace_context,
            "wandb_api_context": wandb_ctx,
            "cache_prefix_context": cache_prefix_context,
        }


# ServerRequest represents a request object sent to the server
@dataclasses.dataclass
class ServerRequest:
    client_id: str
    name: str
    args: typing.Tuple
    context: ServerRequestContext
    id: int = 0

    @classmethod
    def from_json(cls, json: typing.Any) -> "ServerRequest":
        json["context"] = ServerRequestContext.from_json(json["context"])
        return cls(**json)

    def to_json(self) -> typing.Any:
        return {
            "name": self.name,
            "args": self.args,
            "context": self.context.to_json(),
            "id": self.id,
        }

    def error_response(
        self, http_error_code: int, error: Exception
    ) -> "ServerResponse":
        return ServerResponse(
            http_error_code=http_error_code,
            client_id=self.client_id,
            id=self.id,
            # TODO(DG): this is a hack, we should be able to serialize the exception
            value=str(error),
            error=True,
        )

    def success_response(self, value: typing.Any) -> "ServerResponse":
        return ServerResponse(
            http_error_code=200,
            error=False,
            client_id=self.client_id,
            id=self.id,
            value=value,
        )


# ServerResponse represents a response object returned by the server
@dataclasses.dataclass
class ServerResponse:
    client_id: str
    id: int
    value: typing.Any
    error: bool = False
    http_error_code: typing.Optional[int] = None
    http_error_message: typing.Optional[str] = None

    @classmethod
    def from_json(cls, json: typing.Any) -> "ServerResponse":
        return cls(**json)

    def to_json(self) -> typing.Any:
        return {
            "id": self.id,
            "value": self.value,
            "error": self.error,
            "http_error_code": self.http_error_code,
        }


class ShutDown:
    def __eq__(self, other: Any) -> bool:
        return isinstance(other, ShutDown)


shutdown_request = ServerRequest(
    "", "shutdown", (), ServerRequestContext(None, None, None)
)
shutdown_response = ServerResponse("", 0, ShutDown())

HandlerFunction = Callable[..., Any]


class HandlerNotFoundException(Exception):
    pass


# Server class is responsible for managing server lifecycle and handling requests
class Server:
    def __init__(
        self,
        process: bool = False,
    ) -> None:
        self.handlers: Dict[str, HandlerFunction] = {}

        self.request_handler: typing.Union[threading.Thread, aioprocessing.AioProcess]
        self.request_queue: async_queue.Queue[ServerRequest]

        # The internal response queue is used to communicate results back between the server
        # process and the user process. The server process puts responses into this queue,
        # and then the Queue feeder thread puts them into the appropriate client response queue.
        self._internal_response_queue: async_queue.Queue[ServerResponse]

        # just using a ThreadQueue here since this is for communication between two threads
        # that are both in the user process.
        self.client_response_queues: Dict[
            str,
            typing.Union[
                queue.Queue[ServerResponse], async_queue.ThreadQueue[ServerResponse]
            ],
        ] = {}

        self._shutting_down = multiprocessing.Event()
        self._shutdown_lock = multiprocessing.Lock()

        self._request_handler_ready_event = multiprocessing.Event()
        self._request_handler_ready_to_shut_down_event = multiprocessing.Event()

        self._response_queue_feeder_ready_event = threading.Event()
        self._response_queue_feeder_ready_to_shut_down_event = threading.Event()

        # Register handlers
        self.register_handler_fn("ensure_manifest", self.handle_ensure_manifest)
        self.register_handler_fn(
            "ensure_file_downloaded", self.handle_ensure_file_downloaded
        )
        self.register_handler_fn("ensure_file", self.handle_ensure_file)
        self.register_handler_fn("direct_url", self.handle_direct_url)
        self.register_handler_fn("sleep", self.handle_sleep)

        if process:
            self.request_handler = aioprocessing.AioProcess(
                target=self._request_handler_fn, name="IO Server", daemon=True
            )
            self.request_queue = async_queue.ProcessQueue()
            self._internal_response_queue = async_queue.ProcessQueue()
        else:
            self.request_handler = threading.Thread(
                target=self._request_handler_fn, name="IO Server", daemon=True
            )
            self.request_queue = async_queue.ThreadQueue()
            self._internal_response_queue = async_queue.ThreadQueue()

        # runs in the user process and puts responses from the server into the appropriate
        # client-consumed response queues.
        self.response_queue_router = threading.Thread(
            target=self._response_queue_router_fn, daemon=True
        )

    # server_process runs the server's main coroutine
    def _request_handler_fn(self) -> None:
        try:
            asyncio.run(self._request_handler_fn_main(), debug=True)
        except Exception as e:
            logging.exception(f"Error in request handler process: {e}")
            raise e

    # start starts the server thread or process
    def start(self) -> None:
        self.request_handler.start()
        self.response_queue_router.start()
        self._request_handler_ready_event.wait()
        self._response_queue_feeder_ready_event.wait()
        atexit.register(self.shutdown)

    # cleanup performs cleanup actions, such as flushing stats
    def cleanup(self) -> None:
        statsd.flush()

    # shutdown stops the server and joins the thread/process
    def shutdown(self) -> None:
        with self._shutdown_lock:
            self._shutting_down.set()

            # tell the two auxiliary processes to shutdown
            # (the server process and the response queue feeder thread)
            self.request_queue.put(shutdown_request)
            self._internal_response_queue.put(shutdown_response)

            self._response_queue_feeder_ready_to_shut_down_event.wait()
            self._request_handler_ready_to_shut_down_event.wait()

            self.response_queue_router.join()
            self.request_handler.join()
            self.cleanup()

    def _response_queue_router_fn(self) -> None:
        try:
            asyncio.run(self._response_queue_router_fn_main(), debug=True)
        except Exception as e:
            logging.exception(f"Error in response queue router: {e}")
            raise e

    async def _response_queue_router_fn_main(self) -> None:
        self._response_queue_feeder_ready_event.set()
        while True:
            try:
                resp = await self._internal_response_queue.async_get()
            except RuntimeError:
                # this happens when the interpreter is shutting down
                break
            if resp.value == ShutDown():
                self._internal_response_queue.task_done()
                self._internal_response_queue.join()
                break
            client_response_queue = self.client_response_queues[resp.client_id]
            # this is non-blocking b/c resp is already in memory
            client_response_queue.put(resp)
            self._internal_response_queue.task_done()
        # drain queue
        self._response_queue_feeder_ready_to_shut_down_event.set()

    async def _handle(self, req: ServerRequest) -> None:
        with tracer.trace("WBArtifactManager.handle.%s" % req.name, service="weave-am"):
            try:
                handler = self.handlers[req.name]
            except KeyError as e:
                resp = req.error_response(404, e)
            else:
                try:
                    val = await handler(*req.args)
                except Exception as e:
                    logging.error(
                        "WBArtifactManager request error: %s\n",
                        traceback.format_exc(),
                    )
                    print(
                        "WBArtifactManager request error: %s\n",
                        traceback.format_exc(),
                    )
                    resp = req.error_response(
                        server_error_handling.maybe_extract_code_from_exception(e)
                        or 500,
                        e,
                    )
                else:
                    resp = req.success_response(val)
            if isinstance(self._internal_response_queue, async_queue.ThreadQueue):
                # this is fast
                self._internal_response_queue.put(resp)
            else:
                # this needs to perform IPC
                await self._internal_response_queue.async_put(resp)

    # main is the server's main coroutine, handling incoming requests
    async def _request_handler_fn_main(self) -> None:
        # This loop should never block.
        fs = filesystem.FilesystemAsync()
        net = weave_http.HttpAsync(fs)
        loop = asyncio.get_running_loop()
        active_tasks: set[asyncio.Task[typing.Any]] = set()
        async with net:
            self.wandb_file_manager = wandb_file_manager.WandbFileManagerAsync(
                fs, net, await wandb_api.get_wandb_api()
            )

            self._request_handler_ready_event.set()
            while True:
                try:
                    req = await self.request_queue.async_get()
                except RuntimeError as e:
                    print("IO SERVICE RuntimeError", e)
                    # this happens when the loop is shutting down
                    break
                if req.name == "shutdown":
                    self.request_queue.task_done()
                    self.request_queue.join()
                    break
                tracer.context_provider.activate(req.context.trace_context)
                with wandb_api.wandb_api_context(req.context.wandb_api_context):
                    with cache.time_interval_cache_prefix(
                        req.context.cache_prefix_context
                    ):
                        # launch a task to handle the request
                        task = loop.create_task(self._handle(req))
                        active_tasks.add(task)
                        task.add_done_callback(active_tasks.discard)
                self.request_queue.task_done()
        self._request_handler_ready_to_shut_down_event.set()

    def register_handler_fn(self, name: str, handler: HandlerFunction) -> None:
        self.handlers[name] = handler

    @contextlib.contextmanager
    def registered_client(
        self, client: typing.Union["AsyncClient", "SyncClient"]
    ) -> Iterator[None]:
        self.register_client(client)
        try:
            yield
        finally:
            self.unregister_client(client)

    def register_client(
        self, client: typing.Union["SyncClient", "AsyncClient"]
    ) -> None:
        if client.client_id not in self.client_response_queues:
            if isinstance(client, SyncClient):
                self.client_response_queues[client.client_id] = queue.Queue()
            else:
                self.client_response_queues[
                    client.client_id
                ] = async_queue.ThreadQueue()

    def unregister_client(
        self, client: typing.Union["SyncClient", "AsyncClient"]
    ) -> None:
        if client.client_id in self.client_response_queues:
            del self.client_response_queues[client.client_id]

    async def handle_ensure_manifest(
        self, artifact_uri: str
    ) -> typing.Optional[artifact_wandb.WandbArtifactManifest]:
        uri = uris.WeaveURI.parse(artifact_uri)
        if not isinstance(
            uri,
            (artifact_wandb.WeaveWBArtifactURI, artifact_wandb.WeaveWBArtifactByIDURI),
        ):
            raise errors.WeaveInternalError("invalid scheme ", uri)
        return await self.wandb_file_manager.manifest(uri)

    async def handle_ensure_file(self, artifact_uri: str) -> typing.Optional[str]:
        uri = uris.WeaveURI.parse(artifact_uri)
        if not isinstance(
            uri,
            (artifact_wandb.WeaveWBArtifactURI, artifact_wandb.WeaveWBArtifactByIDURI),
        ):
            raise errors.WeaveInternalError("invalid scheme ", uri)
        return await self.wandb_file_manager.ensure_file(uri)

    async def handle_ensure_file_downloaded(
        self, download_url: str
    ) -> typing.Optional[str]:
        return await self.wandb_file_manager.ensure_file_downloaded(download_url)

    async def handle_direct_url(self, artifact_uri: str) -> typing.Optional[str]:
        uri = uris.WeaveURI.parse(artifact_uri)
        if not isinstance(
            uri,
            (artifact_wandb.WeaveWBArtifactURI, artifact_wandb.WeaveWBArtifactByIDURI),
        ):
            raise errors.WeaveInternalError("invalid scheme ", uri)
        return await self.wandb_file_manager.direct_url(uri)

    async def handle_sleep(self, seconds: float) -> float:
        # used for testing to simulate long running processes
        await asyncio.sleep(seconds)
        return seconds


SERVER = None
SERVER_START_LOCK = threading.Lock()


def get_server() -> Server:
    global SERVER
    with SERVER_START_LOCK:
        if SERVER is None:
            SERVER = Server(process=False)
            SERVER.start()
        return SERVER


class AsyncConnection:
    def __init__(
        self,
        client_id: str,
        server: Server,
    ) -> None:
        self.client_id = client_id
        self.server = server
        response_queue = server.client_response_queues[client_id]
        request_queue = server.request_queue

        self.request_id = 0
        self.requests: typing.Dict[int, asyncio.Future] = {}
        self.request_queue: async_queue.Queue[ServerRequest] = request_queue
        self.response_queue: async_queue.ThreadQueue[ServerResponse] = typing.cast(
            async_queue.ThreadQueue[ServerResponse], response_queue
        )
        self.response_task = asyncio.create_task(self.handle_responses())
        self.response_task.add_done_callback(self.response_task_ended)
        self.connected = True

    def response_task_ended(self, task: asyncio.Task) -> None:
        exc = task.exception()
        if exc:
            print("IOServiceError", exc)
            traceback.print_exception(type(exc), exc, exc.__traceback__)
            raise exc

    async def close(self) -> None:
        self.connected = False

    async def handle_responses(self) -> None:
        while self.connected:
            resp = await self.response_queue.async_get()
            self.response_queue.task_done()
            self.requests[resp.id].set_result(resp)

    async def request(self, name: str, *args: typing.Any) -> typing.Any:
        # Caller must check ServerResponse.error!

        wb_ctx = wandb_api.get_wandb_api_context()
        cur_trace_context = tracer.current_trace_context()
        cache_prefix = cache.get_cache_prefix_context()

        req = ServerRequest(
            self.client_id,
            name,
            args,
            ServerRequestContext(cur_trace_context, wb_ctx, cache_prefix),
            self.request_id,
        )

        self.request_id += 1
        response_future: asyncio.Future[ServerResponse] = asyncio.Future()
        self.requests[req.id] = response_future

        with self.server._shutdown_lock:
            is_shutting_down = self.server._shutting_down.is_set()

        if is_shutting_down:
            self.response_task.cancel()
            raise errors.WeaveWandbArtifactManagerError(
                "Server is shutting down, cannot make request"
            )

        if isinstance(self.request_queue, async_queue.ThreadQueue):
            # this is fast
            self.request_queue.put(req)
        else:
            await self.request_queue.async_put(req)
        server_resp = await response_future

        if server_resp.error:
            if server_resp.http_error_code != None:
                raise server_error_handling.WeaveInternalHttpException.from_code(
                    server_resp.http_error_code
                )
            raise errors.WeaveWandbArtifactManagerError(
                "Request error: " + server_resp.value
            )

        return server_resp.value

    async def manifest(
        self,
        artifact_uri: typing.Union[
            artifact_wandb.WeaveWBArtifactURI, artifact_wandb.WeaveWBArtifactByIDURI
        ],
    ) -> typing.Optional[artifact_wandb.WandbArtifactManifest]:
        manifest: typing.Optional[
            artifact_wandb.WandbArtifactManifest
        ] = await self.request("ensure_manifest", str(artifact_uri))
        return manifest

    async def ensure_file(
        self,
        artifact_uri: typing.Union[
            artifact_wandb.WeaveWBArtifactURI, artifact_wandb.WeaveWBArtifactByIDURI
        ],
    ) -> typing.Optional[str]:
        res = await self.request("ensure_file", str(artifact_uri))
        return res

    async def ensure_file_downloaded(self, download_url: str) -> typing.Optional[str]:
        res = await self.request("ensure_file_downloaded", download_url)
        return res

    async def direct_url(
        self,
        artifact_uri: typing.Union[
            artifact_wandb.WeaveWBArtifactURI, artifact_wandb.WeaveWBArtifactByIDURI
        ],
    ) -> typing.Optional[str]:
        res = await self.request("direct_url", str(artifact_uri))
        return res

    async def sleep(self, seconds: float) -> float:
        return await self.request("sleep", seconds)


class AsyncClient:
    def __init__(self, server: Server) -> None:
        self.client_id = str(uuid.uuid4())
        self.server = server

    @contextlib.asynccontextmanager
    async def connect(self) -> typing.AsyncGenerator[AsyncConnection, None]:
        with self.server.registered_client(self):
            conn = AsyncConnection(self.client_id, self.server)
            try:
                yield conn
            finally:
                await conn.close()


class SyncClient:
    def __init__(self, server: Server, fs: filesystem.Filesystem) -> None:
        self.client_id = str(uuid.uuid4())
        self.fs = fs
        self.server = server
        self._current_request_id = 0

    def request(self, name: str, *args: typing.Any) -> typing.Any:
        wb_ctx = wandb_api.get_wandb_api_context()
        cur_trace_context = tracer.current_trace_context()
        cache_prefix = cache.get_cache_prefix_context()
        self._current_request_id += 1

        with self.server.registered_client(self):
            request = ServerRequest(
                self.client_id,
                name,
                args,
                ServerRequestContext(cur_trace_context, wb_ctx, cache_prefix),
                id=self._current_request_id,
            )

            response_queue = self.server.client_response_queues[self.client_id]

            with self.server._shutdown_lock:
                is_shutting_down = self.server._shutting_down.is_set()

            if is_shutting_down:
                raise errors.WeaveWandbArtifactManagerError(
                    "Server is shutting down, cannot make request"
                )

            self.server.request_queue.put(request)
            server_resp = response_queue.get()
            response_queue.task_done()

        if server_resp.error:
            if server_resp.http_error_code != None:
                raise server_error_handling.WeaveInternalHttpException.from_code(
                    server_resp.http_error_code
                )
            raise errors.WeaveWandbArtifactManagerError(
                "Request error: " + server_resp.value
            )
        return server_resp.value

    def manifest(
        self,
        artifact_uri: typing.Union[
            artifact_wandb.WeaveWBArtifactURI, artifact_wandb.WeaveWBArtifactByIDURI
        ],
    ) -> typing.Optional[artifact_wandb.WandbArtifactManifest]:
        manifest: typing.Optional[artifact_wandb.WandbArtifactManifest] = self.request(
            "ensure_manifest", str(artifact_uri)
        )
        return manifest

    def ensure_file(
        self,
        artifact_uri: typing.Union[
            artifact_wandb.WeaveWBArtifactURI, artifact_wandb.WeaveWBArtifactByIDURI
        ],
    ) -> typing.Optional[str]:
        return self.request("ensure_file", str(artifact_uri))

    def ensure_file_downloaded(self, download_url: str) -> typing.Optional[str]:
        return self.request("ensure_file_downloaded", download_url)

    def direct_url(
        self,
        artifact_uri: typing.Union[
            artifact_wandb.WeaveWBArtifactURI, artifact_wandb.WeaveWBArtifactByIDURI
        ],
    ) -> typing.Optional[str]:
        return self.request("direct_url", str(artifact_uri))

    def sleep(self, seconds: float) -> None:
        return self.request("sleep", seconds)


class ServerlessClient:
    def __init__(self, fs: filesystem.Filesystem) -> None:
        self.fs = fs
        self.http = weave_http.Http(self.fs)
        self.wandb_api = wandb_api.WandbApi()
        self.wandb_file_manager = wandb_file_manager.WandbFileManager(
            self.fs, self.http, self.wandb_api
        )

    def manifest(
        self,
        artifact_uri: typing.Union[
            artifact_wandb.WeaveWBArtifactURI, artifact_wandb.WeaveWBArtifactByIDURI
        ],
    ) -> typing.Optional[artifact_wandb.WandbArtifactManifest]:
        return self.wandb_file_manager.manifest(artifact_uri)

    def ensure_file(
        self,
        artifact_uri: typing.Union[
            artifact_wandb.WeaveWBArtifactURI, artifact_wandb.WeaveWBArtifactByIDURI
        ],
    ) -> typing.Optional[str]:
        return self.wandb_file_manager.ensure_file(artifact_uri)

    def ensure_file_downloaded(self, download_url: str) -> typing.Optional[str]:
        return self.wandb_file_manager.ensure_file_downloaded(download_url)

    def direct_url(
        self,
        artifact_uri: typing.Union[
            artifact_wandb.WeaveWBArtifactURI, artifact_wandb.WeaveWBArtifactByIDURI
        ],
    ) -> typing.Optional[str]:
        return self.wandb_file_manager.direct_url(artifact_uri)

    def sleep(self, seconds: float) -> None:
        time.sleep(seconds)


def get_sync_client() -> typing.Union[SyncClient, ServerlessClient]:
    if context_state.serverless_io_service():
        # The io service can't be used during atexit handlers, you get an error like
        # "cannot schedule new futures after shutdown". So it can't really be cleaned up
        # appropriately...
        # This is problematic for users of wandb_stream_table, which tries to flush any
        # outstanding logs during atexit. If any of those require the io service, you get
        # a hang. So we just use the serverless client when setting up weaveflow for now.
        # TODO: this is still an issue for users of streamtable outside of weaveflow, and
        # could be an issue on the weave server if we want clean / flushing shutdowns.
        return ServerlessClient(filesystem.get_filesystem())
    return SyncClient(get_server(), filesystem.get_filesystem())


def get_async_client() -> AsyncClient:
    return AsyncClient(get_server())



# Torch vision datasets
#
# TODO:
#   - Doesn't use the much faster ArrowWeaveList
#   - The patterns here aren't final. Prefer the style used in huggingface/models

import typing
from typing import TypeVar

import PIL.Image
from torchvision import datasets

import weave
from weave.legacy.weave import infer_types
from weave.legacy.weave import panels

ExampleType = TypeVar("ExampleType")


class TypedDictAny(typing.TypedDict):
    pass


class TorchVisionDataset(typing.TypedDict):
    name: str
    description: str
    data: dict[str, list[TypedDictAny]]


def make_torchvision_splits(
    dataset_fn, limit, split_specs, example_keys, example_constructor
) -> dict:
    splits = {}
    for split_name, dataset_fn_kwargs in split_specs:
        raw_data = dataset_fn(".", download=True, **dataset_fn_kwargs)
        examples = []
        for _, row in zip(range(limit), raw_data):
            example = example_constructor(**dict(zip(example_keys, row)))
            examples.append(example)
        splits[split_name] = examples
    return splits


@weave.op(
    input_type={
        "dataset": weave.types.Function(
            {}, infer_types.python_type_to_type(TorchVisionDataset)
        )
    }
)
def torch_vision_dataset_card(dataset) -> panels.Card:
    # TODO: split should be chosen from dropdown rather than via tab.
    #   - or ideally, it should look like one big table with a "split" column.
    dataset_type = dataset.type
    splits_type = dataset_type.property_types["data"]
    split_names = splits_type.property_types.keys()
    split_tabs = []
    for split_name in split_names:
        split_data = dataset["data"][split_name]
        split_tabs.append(
            panels.CardTab(
                name=f"{split_name} split",
                content=panels.LabeledItem(
                    item=split_data, height=500, label=f"{split_name} split"
                ),
            ),
        )
    return panels.Card(
        title=dataset["name"],
        subtitle="",
        content=[
            panels.CardTab(
                name="Overview",
                content=panels.Group(
                    items={
                        "description": panels.LabeledItem(
                            item=dataset["description"], label="Description"
                        ),
                    }
                ),
            ),
            panels.CardTab(
                name="Limitations & Use",
                content=panels.LabeledItem(item="tab2", label="tab2-label"),
            ),
            *split_tabs,
        ],
    )


class ImageLabelExample(typing.TypedDict):
    image: PIL.Image.Image  # TODO: size constraints
    label: int  # TODO: enum?


class MnistSplits(typing.TypedDict):
    train: list[ImageLabelExample]
    test: list[ImageLabelExample]


class MnistDataset(TorchVisionDataset):
    data: MnistSplits  # type: ignore


@weave.op(render_info={"type": "function"})
# TODO: 0-argument functions don't work in WeaveJS at the moment
def mnist(limit: int = -1) -> MnistDataset:
    return MnistDataset(
        name="MNIST",
        description="The famous MNIST dataset",
        data=MnistSplits(  # type: ignore
            make_torchvision_splits(
                datasets.MNIST,
                limit,
                [("train", {"train": True}), ("test", {"train": False})],
                ("image", "label"),
                ImageLabelExample,
            )
        ),
    )


class Food101Splits(typing.TypedDict):
    train: list[ImageLabelExample]
    test: list[ImageLabelExample]


class Food101Dataset(TorchVisionDataset):
    data: Food101Splits  # type: ignore


@weave.op(render_info={"type": "function"})
# TODO: 0-argument functions don't work in WeaveJS at the moment
def food101(limit: int = -1) -> Food101Dataset:
    return Food101Dataset(
        name="Food101",
        description="The Food-101 is a challenging data set of 101 food categories, with 101’000 images. For each class, 250 manually reviewed test images are provided as well as 750 training images. On purpose, the training images were not cleaned, and thus still contain some amount of noise. This comes mostly in the form of intense colors and sometimes wrong labels. All images were rescaled to have a maximum side length of 512 pixels.",
        data=Food101Splits(  # type: ignore
            make_torchvision_splits(
                datasets.Food101,
                limit,
                [("train", {"split": "train"}), ("test", {"split": "test"})],
                ("image", "label"),
                ImageLabelExample,
            )
        ),
    )



import dataclasses
import inspect
import typing

import weave
from weave.legacy.weave import weave_internal
from weave.legacy.weave.ecosystem.wandb import weave_plotly


@weave.type()
class ScatterConfig:
    x_fn: weave.Node[typing.Optional[typing.Any]] = dataclasses.field(
        default_factory=lambda: weave.legacy.weave.graph.VoidNode()
    )
    y_fn: weave.Node[typing.Optional[typing.Any]] = dataclasses.field(
        default_factory=lambda: weave.legacy.weave.graph.VoidNode()
    )
    label_fn: weave.Node[typing.Any] = dataclasses.field(
        default_factory=lambda: weave.legacy.weave.graph.VoidNode()
    )


@weave.type()
class Scatter(weave.Panel):
    id = "Scatter"
    input_node: weave.Node[list[typing.Any]]
    config: typing.Optional[ScatterConfig] = None
    _renderAsPanel: typing.Optional[weave_plotly.PanelPlotly] = None

    @weave.op()
    def initialize(self) -> ScatterConfig:
        input_node = self.input_node
        unnested = weave.legacy.weave.ops.unnest(input_node)
        return ScatterConfig(
            x_fn=weave_internal.define_fn(
                {"item": unnested.type.object_type},
                lambda item: item,  # type: ignore
            ),
            y_fn=weave_internal.define_fn(
                {"item": unnested.type.object_type},
                lambda item: item,  # type: ignore
            ),
            label_fn=weave_internal.define_fn(
                {"item": unnested.type.object_type},
                lambda item: item,  # type: ignore
            ),
        )

    # The config render op. This renders the config editor.
    @weave.op()
    def render_config(self) -> weave.legacy.weave.panels.Group:
        config = typing.cast(ScatterConfig, self.config)
        return weave.legacy.weave.panels.Group(
            items={
                "x_fn": weave.legacy.weave.panels.LabeledItem(
                    label="x", item=weave.legacy.weave.panels.FunctionEditor(config.x_fn)
                ),
                "y_fn": weave.legacy.weave.panels.LabeledItem(
                    label="y", item=weave.legacy.weave.panels.FunctionEditor(config.y_fn)
                ),
                "label_fn": weave.legacy.weave.panels.LabeledItem(
                    label="label",
                    item=weave.legacy.weave.panels.FunctionEditor(config.label_fn),
                ),
            }
        )

    # The render op. This renders the panel.
    @weave.op()
    def render(self) -> weave_plotly.PanelPlotly:
        input_node = self.input_node
        config = typing.cast(ScatterConfig, self.config)
        unnested = weave.legacy.weave.ops.unnest(input_node)
        if (
            not weave.types.optional(weave.types.Float()).assign_type(config.x_fn.type)
            or not weave.types.optional(weave.types.Float()).assign_type(
                config.y_fn.type
            )
            or not weave.types.optional(weave.types.String()).assign_type(
                config.label_fn.type
            )
        ):
            return weave.legacy.weave.panels.PanelHtml(weave.legacy.weave.ops.Html("No data"))  # type: ignore
        if config.label_fn.type == weave.types.Invalid():
            plot_data = unnested.map(
                lambda item: weave.legacy.weave.ops.dict_(x=config.x_fn(item), y=config.y_fn(item))  # type: ignore
            )
        else:
            plot_data = unnested.map(
                lambda item: weave.legacy.weave.ops.dict_(
                    x=config.x_fn(item),
                    y=config.y_fn(item),
                    label=config.label_fn(item),
                )  # type: ignore
            )
        fig = weave_plotly.plotly_scatter(plot_data)
        return weave_plotly.PanelPlotly(fig)

    # This function currently requires a paired output_type implementation in WeaveJS!
    # TODO: Fix
    @weave.op(output_type=lambda input_type: input_type["self"].input_node.output_type)
    def selected(self):
        unnested = weave.legacy.weave.ops.unnest(self.input_node)
        config = self.config
        filtered = unnested.filter(
            lambda item: weave.legacy.weave.ops.Boolean.bool_and(
                weave.legacy.weave.ops.Boolean.bool_and(
                    weave.legacy.weave.ops.Boolean.bool_and(
                        config.x_fn(item)
                        > weave.legacy.weave.ops.TypedDict.pick(
                            self._renderAsPanel.config.selected, "xMin"
                        ),
                        config.x_fn(item)
                        < weave.legacy.weave.ops.TypedDict.pick(
                            self._renderAsPanel.config.selected, "xMax"
                        ),
                    ),
                    config.y_fn(item)
                    > weave.legacy.weave.ops.TypedDict.pick(
                        self._renderAsPanel.config.selected, "yMin"
                    ),
                ),
                config.y_fn(item)
                < weave.legacy.weave.ops.TypedDict.pick(
                    self._renderAsPanel.config.selected, "yMax"
                ),
            )
        )
        return weave_internal.use(filtered)



import re


class RegexMatcher:
    def __init__(self, regex):
        self.regex = regex

    def __eq__(self, other):
        assert isinstance(other, str)
        return bool(re.match(self.regex, other))

    def __repr__(self):
        return "<RegexMatcher '%s'>" % self.regex


def assert_nodes_equal(n1, n2):
    # Don't override __eq__ on node! Nodes are mixed in with a class that
    #     contains weave methods for the given type, some of which need
    #     to override __eq__
    assert n1.to_json() == n2.to_json()



import typing
import urllib

from weave.legacy.weave import errors
from weave.legacy.weave import weave_types as types
from weave.legacy.weave.api import op
from weave.legacy.weave.gql_op_plugin import wb_gql_op_plugin
from weave.legacy.weave.ops_domain import wb_domain_types as wdt
from weave.legacy.weave.ops_domain.wandb_domain_gql import (
    gql_connection_op,
    gql_direct_edge_op,
    gql_prop_op,
    make_root_op_gql_op_output_type,
)

# Section 1/6: Tag Getters
# None


# Section 2/6: Root Ops
@op(
    name="root-allArtifactsGQLResolver",
    input_type={"gql_result": types.TypedDict({})},
    output_type=types.List(wdt.ArtifactCollectionType),
    hidden=True,
)
def root_all_artifacts_gql_resolver(gql_result):
    return [
        wdt.ArtifactCollection.from_keys(artifact_collection["node"])
        for artifact_collection in gql_result["instance"]["artifacts_500"]["edges"]
    ]


@op(
    name="root-allArtifacts",
    input_type={},
    output_type=types.List(wdt.ArtifactCollectionType),
    plugins=wb_gql_op_plugin(
        lambda inputs, inner: f"""
    instance {{
        artifacts_500: artifactSequences(limit: 500) {{
            edges {{
                node {{
                    {wdt.ArtifactCollection.REQUIRED_FRAGMENT}
                    {inner}
                }}
            }}
        }}
    }}
    """,
        is_root=True,
        root_resolver=root_all_artifacts_gql_resolver,
        gql_op_output_type=make_root_op_gql_op_output_type(
            "artifacts_500", lambda inputs: "", wdt.ArtifactCollectionType
        ),
    ),
)
def root_all_artifacts():
    raise errors.WeaveGQLCompileError(
        "root-allArtifacts should not be executed directly. If you see this error, it is a bug in the Weave compiler."
    )


# Section 3/6: Attribute Getters
gql_prop_op(
    "artifact-id",
    wdt.ArtifactCollectionType,
    "id",
    types.String(),
)
artifact_name = gql_prop_op(
    "artifact-name",
    wdt.ArtifactCollectionType,
    "name",
    types.String(),
)
gql_prop_op(
    "artifact-description",
    wdt.ArtifactCollectionType,
    "description",
    types.String(),
)
gql_prop_op(
    "artifact-createdAt",
    wdt.ArtifactCollectionType,
    "createdAt",
    types.Timestamp(),
)

# Section 4/6: Direct Relationship Ops
gql_direct_edge_op(
    "artifact-type",
    wdt.ArtifactCollectionType,
    "defaultArtifactType",
    wdt.ArtifactTypeType,
)
gql_direct_edge_op(
    "artifact-project",
    wdt.ArtifactCollectionType,
    "project",
    wdt.ProjectType,
)

gql_direct_edge_op(
    "artifact-membershipForAlias",
    wdt.ArtifactCollectionType,
    "artifactMembership",
    wdt.ArtifactCollectionMembershipType,
    {
        "aliasName": types.String(),
    },
    lambda inputs: f'aliasName: {inputs["aliasName"]}',
)

# Section 5/6: Connection Ops
gql_connection_op(
    "artifact-versions",
    wdt.ArtifactCollectionType,
    "artifacts",
    wdt.ArtifactVersionType,
    {},
    lambda inputs: "first: 100",
)
gql_connection_op(
    "artifact-memberships",
    wdt.ArtifactCollectionType,
    "artifactMemberships",
    wdt.ArtifactCollectionMembershipType,
    {},
)
gql_connection_op(
    "artifact-aliases",
    wdt.ArtifactCollectionType,
    "aliases",
    wdt.ArtifactAliasType,
    {},
)


# Section 6/6: Non Standard Business Logic Ops
@op(
    name="artifact-isPortfolio",
    plugins=wb_gql_op_plugin(
        lambda inputs, inner: "__typename",
    ),
)
def is_portfolio(artifact: wdt.ArtifactCollection) -> bool:
    return artifact["__typename"] == "ArtifactPortfolio"


@op(
    name="artifact-lastMembership",
    plugins=wb_gql_op_plugin(
        lambda inputs, inner: f"""
            artifactMemberships_first_1: artifactMemberships(first: 1) {{
                edges {{
                    node {{
                        {wdt.ArtifactCollectionMembership.REQUIRED_FRAGMENT}
                        {inner}
                    }}
                }}
            }}
        """,
        gql_op_output_type=lambda inputs, input_type: types.optional(
            wdt.ArtifactCollectionMembershipType.with_keys(
                typing.cast(typing.Any, input_type)
                .keys["artifactMemberships_first_1"]["edges"]
                .object_type["node"]
                .property_types
            )
        ),
    ),
)
def last_membership(
    artifact: wdt.ArtifactCollection,
) -> typing.Optional[wdt.ArtifactCollectionMembership]:
    edges = artifact["artifactMemberships_first_1"]["edges"]
    if len(edges) == 0:
        return None
    return wdt.ArtifactCollectionMembership.from_keys(edges[0]["node"])


@op(
    name="artifact-link",
    plugins=wb_gql_op_plugin(
        lambda inputs, inner: """
            defaultArtifactType {
                id
                name
                project {
                    id
                    name
                    entity {
                        id
                        name
                    }
                }
            }
        """,
    ),
)
def link(
    artifact: wdt.ArtifactCollection,
) -> wdt.Link:
    artifact_type = artifact["defaultArtifactType"]
    project = artifact_type["project"]
    entity = project["entity"]
    entity_name = entity["name"]
    project_name = project["name"]
    artifact_type_name = artifact_type["name"]
    artifact_name = artifact["name"]
    return wdt.Link(
        artifact["name"],
        f"/{entity_name}/{project_name}/artifacts/{urllib.parse.quote(artifact_type_name)}/{urllib.parse.quote(artifact_name)}",
    )


@op(
    name="artifact-rawTags",
    output_type=types.List(
        types.TypedDict(
            {
                "id": types.String(),
                "name": types.String(),
                "tagCategoryName": types.String(),
                "attributes": types.String(),
            }
        )
    ),
    plugins=wb_gql_op_plugin(
        lambda inputs, inner: """
        tags {
            edges {
                node {
                    id
                    name
                    tagCategoryName
                    attributes
                }
            }
        }
        """
    ),
)
def raw_tags(artifact: wdt.ArtifactCollection):
    return [n["node"] for n in artifact["tags"]["edges"]]



import socket
from typing import Optional

import analytics

from weave.legacy.weave import context_state

from . import environment

analytics.write_key = "uJ8vZgKqTBVH6ZdhD4GZGZYsR7ucfJmb"


def hostname() -> str:
    return socket.gethostname()


identify_called = False


def _identify() -> None:
    global identify_called
    if not identify_called:
        host = hostname()
        analytics.identify(host, {"hostname": host})
    identify_called = True


def analytics_enabled() -> bool:
    context_enabled: bool = context_state.analytics_enabled()
    env_enabled: bool = environment.usage_analytics_enabled()

    return context_enabled and env_enabled


def track(action: str, info: Optional[dict] = None) -> None:
    if not analytics_enabled():
        return
    try:
        _identify()
        if info is None:
            info = {}
        analytics.track(hostname(), action, info)
    except:
        pass


def use_called() -> None:
    track("called use")


def show_called(info: Optional[dict] = None) -> None:
    track("called show", info)
